@@ -51,6 +51,7 @@ from jax._src.config import (
   check_tracer_leaks as check_tracer_leaks,
   checking_leaks as checking_leaks,
   enable_custom_prng as enable_custom_prng,
+  softmax_custom_jvp as softmax_custom_jvp,
   enable_custom_vjp_by_custom_transpose as enable_custom_vjp_by_custom_transpose,
   debug_nans as debug_nans,
   debug_infs as debug_infs,
@@ -76,9 +77,10 @@ del _xc
 
 from jax._src.api import effects_barrier as effects_barrier
 from jax._src.api import block_until_ready as block_until_ready
-from jax._src.api import checkpoint as checkpoint
+from jax._src.ad_checkpoint import checkpoint_wrapper as checkpoint
 from jax._src.ad_checkpoint import checkpoint_policies as checkpoint_policies
 from jax._src.api import clear_backends as clear_backends
+from jax._src.api import clear_caches as clear_caches
 from jax._src.custom_derivatives import closure_convert as closure_convert
 from jax._src.util import curry as _deprecated_curry
 from jax._src.custom_derivatives import custom_gradient as custom_gradient
@@ -116,8 +118,8 @@ from jax._src.api import named_scope as named_scope
 from jax._src.api import pmap as pmap
 from jax._src.xla_bridge import process_count as process_count
 from jax._src.xla_bridge import process_index as process_index
-from jax._src.api import pure_callback as pure_callback
-from jax._src.api import remat as remat
+from jax._src.callback import pure_callback_api as pure_callback
+from jax._src.ad_checkpoint import checkpoint_wrapper as remat
 from jax._src.core import ShapedArray as _deprecated_ShapedArray
 from jax._src.api import ShapeDtypeStruct as ShapeDtypeStruct
 from jax._src.api import value_and_grad as value_and_grad
@@ -126,6 +128,8 @@ from jax._src.api import vmap as vmap
 from jax._src.api import xla_computation as xla_computation
 
 from jax.interpreters import ad as _deprecated_ad
+import jax.interpreters.batching
+import jax.interpreters.mlir
 from jax.interpreters import partial_eval as _deprecated_partial_eval
 from jax.interpreters import pxla as _deprecated_pxla
 from jax.interpreters import xla as _deprecated_xla
@@ -151,7 +155,10 @@ from jax._src.tree_util import (
 
 # These submodules are separate because they are in an import cycle with
 # jax and rely on the names imported above.
-from jax import abstract_arrays as abstract_arrays
+from jax import abstract_arrays as _deprecated_abstract_arrays
+from jax import custom_derivatives as custom_derivatives
+from jax import custom_batching as custom_batching
+from jax import custom_transpose as custom_transpose
 from jax import api_util as api_util
 from jax import distributed as distributed
 from jax import debug as debug
@@ -160,6 +167,7 @@ from jax import errors as errors
 from jax import image as image
 from jax import lax as lax
 from jax import linear_util as linear_util
+from jax import monitoring as monitoring
 from jax import nn as nn
 from jax import numpy as numpy
 from jax import ops as ops
@@ -174,8 +182,15 @@ from jax import util as util
 # Also circular dependency.
 from jax._src.array import Shard as Shard
 
+import jax.experimental.compilation_cache.compilation_cache as _ccache
+del _ccache
 
 _deprecations = {
+  # Added 06 June 2023
+  "abstract_arrays": (
+    "jax.abstract_arrays is deprecated. Refer to jax.core.",
+    _deprecated_abstract_arrays
+  ),
   # Added 28 March 2023
   "ShapedArray": (
     "jax.ShapedArray is deprecated. Use jax.core.ShapedArray",
@@ -209,6 +224,7 @@ _deprecations = {
 
 import typing as _typing
 if _typing.TYPE_CHECKING:
+  from jax._src import abstract_arrays as abstract_arrays
   from jax._src.core import ShapedArray as ShapedArray
   from jax.interpreters import ad as ad
   from jax.interpreters import partial_eval as partial_eval
