@@ -11,6 +11,7 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import annotations
 
 import dataclasses
 import functools
@@ -20,31 +21,33 @@ from typing import (Union, Optional, Callable, Dict, Tuple, TypeVar,
 
 import numpy as np
 
-import jax
 import jax.numpy as jnp
-import jax.tree_util as jtu
 from jax import lax
-from jax.api_util import flatten_fun
-from jax.experimental import pjit
-from jax.tree_util import tree_flatten
-from jax.tree_util import tree_map
-from jax.tree_util import tree_unflatten
 
+from jax._src import api
 from jax._src import linear_util as lu
 from jax._src import core
 from jax._src import custom_derivatives
 from jax._src import effects
+from jax._src import pjit
 from jax._src import prng
+from jax._src import sharding_impls
 from jax._src import source_info_util
 from jax._src import traceback_util
+from jax._src import tree_util as jtu
+from jax._src.ad_util import SymbolicZero
+from jax._src.api_util import flatten_fun
 from jax._src.config import config
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
 from jax._src.interpreters import partial_eval as pe
+from jax._src.tree_util import tree_flatten
+from jax._src.tree_util import tree_map
+from jax._src.tree_util import tree_unflatten
 from jax._src.typing import Array
 from jax._src.util import (as_hashable_function, split_list, safe_map, safe_zip,
-                           unzip3, weakref_lru_cache)
+                           unzip3, weakref_lru_cache, HashableWrapper)
 
 source_info_util.register_exclusion(__file__)
 traceback_util.register_exclusion(__file__)
@@ -92,7 +95,7 @@ class JaxException(Exception):
     del payload
     return cls(metadata)
 
-  def get_effect_type(self) -> core.Effect:
+  def get_effect_type(self) -> ErrorEffect:
     raise NotImplementedError
 
 
@@ -100,7 +103,7 @@ class JaxException(Exception):
 @dataclasses.dataclass(eq=True, frozen=True)
 class ErrorEffect(effects.Effect):
   error_type: Type[JaxException]
-  shape_dtypes: Tuple[jax.ShapeDtypeStruct, ...]
+  shape_dtypes: Tuple[api.ShapeDtypeStruct, ...]
 
   def __lt__(self, other: 'ErrorEffect'):
     shape_dtypes = lambda x: tuple((sd.shape, str(sd.dtype))  # dtype is not comparable
@@ -114,7 +117,7 @@ effects.lowerable_effects.add_type(ErrorEffect)
 class DivisionByZeroError(JaxException):
 
   def __str__(self):
-    return f'division by zero at {self.traceback_info}'
+    return f'division by zero at:\n\n{self.traceback_info}'
 
   def get_effect_type(self):
     return ErrorEffect(DivisionByZeroError, ())
@@ -136,7 +139,7 @@ class NaNError(JaxException):
     return ErrorEffect(NaNError, ())
 
   def __str__(self):
-    return f'nan generated by primitive: {self.prim} at {self.traceback_info}'
+    return f'nan generated by primitive: {self.prim} at:\n\n{self.traceback_info}'
 
 class OOBError(JaxException):
 
@@ -158,10 +161,10 @@ class OOBError(JaxException):
             f'shape {self.operand_shape}: '
             f'index {self._payload[0]} is out of bounds for axis '
             f'{self._payload[1]} with size {self._payload[2]}. '
-            f'Failed at {self.traceback_info}')
+            f'Failed at:\n\n{self.traceback_info}')
 
   def get_effect_type(self):
-    return ErrorEffect(OOBError, (jax.ShapeDtypeStruct((3,), jnp.int32),))
+    return ErrorEffect(OOBError, (api.ShapeDtypeStruct((3,), jnp.int32),))
 
 class FailedCheckError(JaxException):
 
@@ -182,13 +185,13 @@ class FailedCheckError(JaxException):
 
   def __str__(self):
     return (self.fmt_string.format(*self.args, **self.kwargs)
-            + f' (check failed at {self.traceback_info})')
+            + f' `check` failed at:\n\n{self.traceback_info}')
 
   def get_effect_type(self):
     vals = jtu.tree_leaves((self.args, self.kwargs))
     return ErrorEffect(
         FailedCheckError,
-        tuple(jax.ShapeDtypeStruct(x.shape, x.dtype) for x in vals))
+        tuple(api.ShapeDtypeStruct(x.shape, x.dtype) for x in vals))
 
 @dataclasses.dataclass
 class BatchedError(JaxException):
@@ -350,8 +353,13 @@ def default_checkify_rule(primitive: core.Primitive, error: Error,
 
   # call_jaxpr handling
   call_jaxpr = params.pop('call_jaxpr')
+  if isinstance(call_jaxpr, core.ClosedJaxpr):  # handle closed_call_p
+    jaxpr, consts = call_jaxpr.jaxpr, call_jaxpr.consts
+  else:
+    jaxpr, consts = call_jaxpr, ()
+  consts_ = tuple(HashableWrapper(c) for c in consts)
   partial_checkify = lu.hashable_partial(lu.wrap_init(
-      checkify_jaxpr_flat), call_jaxpr, (), enabled_errors, err_tree)
+      checkify_jaxpr_flat_hashable), jaxpr, consts_, enabled_errors, err_tree)
   partial_checkify, metadata = _flatten_and_get_error_metadata_thunk(
       partial_checkify)
 
@@ -380,20 +388,6 @@ def default_checkify_rule(primitive: core.Primitive, error: Error,
 def get_shaped_aval(val):
   return core.raise_to_shaped(core.get_aval(val))
 
-def initial_style_jaxpr(
-    fun: Callable, in_tree: PyTreeDef, in_avals: Sequence[core.AbstractValue]
-  ) -> Tuple[core.Jaxpr, List[Any], PyTreeDef]:
-  return _initial_style_jaxpr(fun, in_tree, tuple(in_avals))
-
-@weakref_lru_cache
-def _initial_style_jaxpr(fun, in_tree, in_avals):
-  # like control_flow._initial_style_jaxpr, but use flatten_fun not _nokwargs
-  fun_, out_tree = flatten_fun(lu.wrap_init(fun), in_tree)
-  debug = pe.debug_info(fun, in_tree, out_tree, False, 'checkify')
-  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(fun_, in_avals, debug)
-  return jaxpr, consts, out_tree()
-
-
 def checkify_jaxpr(jaxpr: core.ClosedJaxpr, enabled_errors,
                    error: Error, *args) -> Tuple[Error, List[core.Value]]:
   err_vals, err_tree = jtu.tree_flatten(error)
@@ -435,6 +429,11 @@ def checkify_jaxpr_flat(jaxpr: core.Jaxpr, consts: Sequence[core.Value],
 
   return error, map(read_env, jaxpr.outvars)
 
+def checkify_jaxpr_flat_hashable(jaxpr, hashable_consts, enabled_errors,
+                                 err_tree, *args):
+  consts = tuple(c.x for c in hashable_consts)
+  return checkify_jaxpr_flat(jaxpr, consts, enabled_errors, err_tree, *args)
+
 @lu.transformation_with_aux
 def flatten_fun_output(*args):
   ans = yield args, {}
@@ -545,7 +544,9 @@ error_checks: Dict[core.Primitive, ErrorCheckRule] = {}
 
 
 def summary() -> str:
-  return str(source_info_util.summarize(source_info_util.current()))
+  return str(source_info_util.summarize(
+      source_info_util.current(),
+      config.read("jax_tracer_error_num_traceback_frames")))
 
 def nan_error_check(prim, error, enabled_errors, *in_vals, **params):
   out = prim.bind(*in_vals, **params)
@@ -587,6 +588,38 @@ for _prim in nan_primitives:
   error_checks[_prim] = functools.partial(nan_error_check, _prim)
 
 
+def dynamic_slice_error_check(error, enabled_errors, operand, *start_indices, slice_sizes):
+  out = lax.dynamic_slice_p.bind(operand, *start_indices, slice_sizes=slice_sizes)
+
+  if OOBError not in enabled_errors:
+    return error, out
+
+  operand_dims = np.array(operand.shape)
+  slice_sizes = np.array(slice_sizes)
+  start_indices = jnp.array(start_indices)
+  oob_mask = (start_indices < 0) | (start_indices + slice_sizes > operand_dims)
+
+  payload = oob_payload(oob_mask, start_indices, range(operand.ndim), operand.shape)
+  error = assert_func(error, jnp.any(oob_mask), OOBError(summary(), "dynamic_slice", operand.shape, payload))
+  return error, out
+error_checks[lax.dynamic_slice_p] = dynamic_slice_error_check
+
+def dynamic_update_slice_error_check(error, enabled_errors, operand, update, *start_indices):
+  out = lax.dynamic_update_slice_p.bind(operand, update, *start_indices)
+
+  if OOBError not in enabled_errors:
+    return error, out
+
+  operand_dims = np.array(operand.shape)
+  update_dims = np.array(update.shape)
+  start_indices = jnp.array(start_indices)
+  oob_mask = (start_indices < 0) | (start_indices + update_dims > operand_dims)
+
+  payload = oob_payload(oob_mask, start_indices, range(operand.ndim), operand.shape)
+  error = assert_func(error, jnp.any(oob_mask), OOBError(summary(), "dynamic_update_slice", operand.shape, payload))
+  return error, out
+error_checks[lax.dynamic_update_slice_p] = dynamic_update_slice_error_check
+
 def gather_error_check(error, enabled_errors, operand, start_indices, *,
                        dimension_numbers, slice_sizes, unique_indices,
                        indices_are_sorted, mode, fill_value):
@@ -772,20 +805,23 @@ error_checks[lax.scan_p] = scan_error_check
 def checkify_while_body_jaxpr(
     cond_jaxpr: core.ClosedJaxpr, body_jaxpr: core.ClosedJaxpr,
     enabled_errors, error: Error,
-    c_consts) -> Tuple[core.ClosedJaxpr, PyTreeDef, Set[ErrorEffect]]:
+    c_consts_num: int) -> Tuple[core.ClosedJaxpr, PyTreeDef, Set[ErrorEffect]]:
   cond_f = core.jaxpr_as_fun(cond_jaxpr)
   body_f = core.jaxpr_as_fun(body_jaxpr)
-  def new_body_f(*vals):
+  def new_body_f(*c_consts_and_vals):
+    c_consts, vals = split_list(c_consts_and_vals, [c_consts_num])
     out = body_f(*vals)
     # This checks if the next cond application will error
     _ = cond_f(*c_consts, *out)
     return out
   new_body_f_ = lu.wrap_init(new_body_f)
-  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(new_body_f_, body_jaxpr.in_avals)
-  closed_jaxpr = core.ClosedJaxpr(jaxpr, consts)
+  c_consts_avals = cond_jaxpr.in_avals[:c_consts_num]
+  jaxpr, _, () = pe.trace_to_jaxpr_dynamic(new_body_f_, [*c_consts_avals,
+                                                         *body_jaxpr.in_avals])
+  closed_jaxpr = core.ClosedJaxpr(jaxpr, ())
   err_vals, err_tree = jtu.tree_flatten(error)
   err_vals = map(get_shaped_aval, err_vals)
-  flat_err_and_in_vals = [*err_vals, *body_jaxpr.in_avals]
+  flat_err_and_in_vals = [*err_vals, *c_consts_avals, *body_jaxpr.in_avals]
   jaxpr, out_tree, error_effects = jaxpr_to_checkify_jaxpr(
       closed_jaxpr, enabled_errors, err_tree, *flat_err_and_in_vals)
   return jaxpr, out_tree, error_effects
@@ -811,14 +847,16 @@ def while_loop_error_check(error, enabled_errors, *in_flat, cond_nconsts,
   error, _ = checkify_jaxpr(cond_jaxpr, enabled_errors, error, *c_consts, *carry)
 
   _, _, error_effects = checkify_while_body_jaxpr(cond_jaxpr, body_jaxpr,
-                                                  enabled_errors, error, c_consts)
+                                                  enabled_errors, error,
+                                                  cond_nconsts)
   # merged error!
   error = error._add_placeholder_effects(error_effects)
   err_vals, err_tree = jtu.tree_flatten(error)
   checked_body_jaxpr_, body_out_tree, _ = checkify_while_body_jaxpr(
-      cond_jaxpr, body_jaxpr, enabled_errors, error, c_consts)
+      cond_jaxpr, body_jaxpr, enabled_errors, error, cond_nconsts)
   num_error_vals = len(err_vals)
-  to_move = [False] * num_error_vals + [True] * body_nconsts + [False] * len(carry)
+  to_move = ([False] * num_error_vals + [True] * cond_nconsts
+             + [True] * body_nconsts + [False] * len(carry))
   checked_body_jaxpr = pe.move_binders_to_front(checked_body_jaxpr_, to_move)
 
   cond_in_flat = [*err_vals, *c_consts, *carry]
@@ -829,10 +867,10 @@ def while_loop_error_check(error, enabled_errors, *in_flat, cond_nconsts,
   to_move = [False] * num_error_vals + [True] * cond_nconsts + [False] * len(carry)
   compat_cond_jaxpr = pe.move_binders_to_front(compat_cond_jaxpr_, to_move)
 
-  new_in_flat = [*c_consts, *b_consts, *err_vals, *carry]
+  new_in_flat = [*c_consts, *c_consts, *b_consts, *err_vals, *carry]
   all_out_vals = lax.while_p.bind(
       *new_in_flat, cond_nconsts=cond_nconsts, cond_jaxpr=compat_cond_jaxpr,
-      body_nconsts=body_nconsts, body_jaxpr=checked_body_jaxpr)
+      body_nconsts=cond_nconsts+body_nconsts, body_jaxpr=checked_body_jaxpr)
   # body_out_tree will have all the metadata of cond because it executes a cond!
   error, out = tree_unflatten(body_out_tree, all_out_vals)
   return error, out
@@ -852,7 +890,7 @@ def pjit_error_check(error, enabled_errors, *vals_in, jaxpr,
   # Update pjit params to account for extra error values.
   num_error_vals = len(err_vals)
   num_out_error_vals = out_tree.num_leaves - len(out_shardings)
-  sharding = pjit._UNSPECIFIED
+  sharding = sharding_impls.UNSPECIFIED
 
   new_in_shardings = (*[sharding] * num_error_vals, *in_shardings)
   new_out_shardings = (*[sharding] * num_out_error_vals, *out_shardings)
@@ -889,17 +927,7 @@ def custom_jvp_call_rule(in_err, enabled_errors, *in_vals, num_consts,
                         call_jaxpr.consts, enabled_errors, err_tree))
   partial_checkify, f_metadata = _flatten_and_get_error_metadata_thunk(
       partial_checkify)
-
-  # Construct the default jvp function, without checkify-ing.
-  @lu.wrap_init
-  def jvp(*xs):
-    # TODO(lenamartens, sharadmv): why not checkify here?
-    jvp_jaxpr, jvp_consts = jvp_jaxpr_thunk()
-    n, ragged = divmod(len(xs), 2)
-    assert not ragged
-    primals, tangents = xs[num_consts:n], xs[n+num_consts:]
-    return core.eval_jaxpr(jvp_jaxpr, jvp_consts, *primals, *tangents)
-
+  jvp = lift_jvp(err_tree.num_leaves, num_consts, jvp_jaxpr_thunk)
   jvp, jvp_out_tree = flatten_fun_output(jvp)
   all_outs = custom_derivatives.custom_jvp_call_p.bind(
       partial_checkify, jvp, *err_vals, *in_vals, **params)
@@ -914,35 +942,67 @@ def custom_jvp_call_rule(in_err, enabled_errors, *in_vals, num_consts,
   return out_err, out_vals
 error_checks[custom_derivatives.custom_jvp_call_p] = custom_jvp_call_rule
 
+# Compared to custom_derivatives.lift_jvp, we're handling the extra inputs and
+# outputs that checkify adds (just forwarding the error data's primal and
+# tangent components). The jaxpr in jvp_jaxpr_thunk doesn't expect those.
+# TODO(mattjj): can we simplify this, or dedup with custom_derivatives.lift_jvp?
+# Adding another layer of lu.transformation was tricky, though maybe doable.
+def lift_jvp(num_errs, num_consts, jvp_jaxpr_thunk):
+  @lu.wrap_init
+  def jvp(*xs):
+    n, ragged = divmod(len(xs), 2)
+    assert not ragged
+    primals, tangents = xs[num_consts+num_errs:n], xs[n+num_consts+num_errs:]
+    zeros = [type(t) is SymbolicZero for t in tangents]
+    jvp_jaxpr, jvp_consts, out_zeros = jvp_jaxpr_thunk(*zeros)
+    nonzero_tangents = [t for t in tangents if type(t) is not SymbolicZero]
+    out = core.eval_jaxpr(jvp_jaxpr, jvp_consts, *primals, *nonzero_tangents)
+    out_primals, nz_out_tangents = split_list(out, [len(out_zeros)])
+    nz_out_tangents_ = iter(nz_out_tangents)
+    out_tangents = [SymbolicZero(core.get_aval(p).at_least_vspace())
+                    if z else next(nz_out_tangents_)
+                    for p, z in zip(out_primals, out_zeros)]
+    assert next(nz_out_tangents_, None) is None
+    primal_errs = xs[num_consts:num_consts+num_errs]
+    tangent_errs = xs[n+num_consts:n+num_consts+num_errs]
+    return [*primal_errs, *out_primals, *tangent_errs, *out_tangents]
+  return jvp
+
 def custom_vjp_call_jaxpr_rule(in_err, enabled_errors, *in_vals, fun_jaxpr,
-                               fwd_jaxpr_thunk, num_consts, bwd, out_trees):
+                               fwd_jaxpr_thunk, num_consts, bwd, out_trees,
+                               symbolic_zeros):
   err_vals, err_tree = jtu.tree_flatten(in_err)
-  fun = lu.wrap_init(
+  num_errs = err_tree.num_leaves
+  checkified_fun = lu.wrap_init(
       functools.partial(checkify_jaxpr_flat, fun_jaxpr.jaxpr,
                         fun_jaxpr.consts, enabled_errors, err_tree))
-  fun, fun_metadata = _flatten_and_get_error_metadata_thunk(fun)
+  checkified_fun, fun_metadata = _flatten_and_get_error_metadata_thunk(
+      checkified_fun)
 
   @lu.wrap_init
-  def fwd(*xs):
+  def checkified_fwd(*args):
     # TODO(lenamartens, sharadmv): why not checkify here?
-    fwd_jaxpr, fwd_consts = fwd_jaxpr_thunk()
+    xs, zeros = args[::2], args[1::2]
+    xs, zeros = xs[num_errs:], zeros[num_errs:]
+    fwd_jaxpr, fwd_consts = fwd_jaxpr_thunk(*zeros)
     xs_without_consts = xs[num_consts:]
     return core.eval_jaxpr(fwd_jaxpr, fwd_consts, *xs_without_consts)
 
-  fwd, fwd_out_tree = flatten_fun_output(fwd)
+  bwd_ = lambda *args: (*(None,)*num_errs, *bwd(*args))
+  checkified_fwd, fwd_out_tree = flatten_fun_output(checkified_fwd)
   all_outs = custom_derivatives.custom_vjp_call_p.bind(
-      fun, fwd, bwd, *err_vals, *in_vals, out_trees=out_trees)
+      checkified_fun, checkified_fwd, bwd_, *err_vals, *in_vals, out_trees=out_trees,
+      symbolic_zeros=symbolic_zeros)
   fst, out_metadata = lu.merge_linear_aux(fun_metadata, fwd_out_tree)
   if fst:
     err_and_out_tree, _ = out_metadata
     out_err, out_vals = tree_unflatten(err_and_out_tree, all_outs)
   else:
-    err_vals, out_vals = split_list(all_outs, [len(err_vals)])
-    # forward input error to output
-    out_err = jtu.tree_unflatten(err_tree, err_vals)
+    out_err, out_vals = in_err, all_outs
   return out_err, out_vals
 error_checks[custom_derivatives.custom_vjp_call_jaxpr_p] = custom_vjp_call_jaxpr_rule
 
+
 def check_discharge_rule(error, enabled_errors, *args, err_tree, debug):
   del debug
   new_error = tree_unflatten(err_tree, args)
@@ -1044,16 +1104,17 @@ def checkify(f: Callable[..., Out],
   """
   @traceback_util.api_boundary
   def checked_fun(*args, **kwargs):
+    # close over all arguments so they're not turned into abstract values.
+    in_tree = jtu.tree_structure(((), {}))
+    closed_f = lambda: f(*args, **kwargs)
     # stage:
-    flat_args, in_tree = tree_flatten((args, kwargs))
-    in_avals = map(get_shaped_aval, flat_args)
-    jaxpr_, consts, out_tree = initial_style_jaxpr(f, in_tree, in_avals)
+    fun_, out_tree = flatten_fun(lu.wrap_init(closed_f), in_tree)
+    debug = pe.debug_info(closed_f, in_tree, out_tree, False, 'checkify')
+    jaxpr_, _, consts = pe.trace_to_jaxpr_dynamic(fun_, (), debug)
     jaxpr = pe.close_jaxpr(pe.convert_constvars_jaxpr(jaxpr_))
     # checkify:
-    flat_args = jtu.tree_leaves((args, kwargs))
-    error, out_flat = checkify_jaxpr(jaxpr, errors, init_error,
-                                     *consts, *flat_args)
-    return error, jtu.tree_unflatten(out_tree, out_flat)
+    error, out_flat = checkify_jaxpr(jaxpr, errors, init_error, *consts)
+    return error, jtu.tree_unflatten(out_tree(), out_flat)
   return checked_fun
 
 def check(pred: Bool, msg: str, *fmt_args, **fmt_kwargs) -> None:
@@ -1096,7 +1157,7 @@ def _check(pred, msg, debug, *fmt_args, **fmt_kwargs):
     prim_name = 'debug_check' if debug else 'check'
     raise TypeError(f'{prim_name} takes a scalar pred as argument, got {pred}')
   for arg in jtu.tree_leaves((fmt_args, fmt_kwargs)):
-    if not isinstance(arg, (jax.Array, np.ndarray)):
+    if not isinstance(arg, (Array, np.ndarray)):
       raise TypeError('Formatting arguments to checkify.check need to be '
                       'PyTrees of arrays, but got '
                       f'{repr(arg)} of type {type(arg)}.')
@@ -1114,7 +1175,7 @@ def _check_error(error, *, debug=False):
 
 def is_scalar_pred(pred) -> bool:
   return (isinstance(pred, bool) or
-          isinstance(pred, jax.Array) and pred.shape == () and
+          isinstance(pred, Array) and pred.shape == () and
           pred.dtype == jnp.dtype('bool'))
 
 
