@@ -14,7 +14,7 @@
 
 import os
 
-running_in_cloud_tpu_vm = False
+running_in_cloud_tpu_vm: bool = False
 
 
 def maybe_import_libtpu():
@@ -35,7 +35,7 @@ def jax_force_tpu_init() -> bool:
   return 'JAX_FORCE_TPU_INIT' in os.environ
 
 
-def cloud_tpu_init():
+def cloud_tpu_init() -> None:
   """Automatically sets Cloud TPU topology and other env vars.
 
   **This must be called before the TPU runtime is loaded, which happens as soon
@@ -67,3 +67,6 @@ def cloud_tpu_init():
   os.environ.setdefault('GRPC_VERBOSITY', 'ERROR')
   os.environ.setdefault('JAX_PLATFORMS', 'tpu,cpu')
   os.environ['TPU_ML_PLATFORM'] = 'JAX'
+
+  if 'JAX_USE_PJRT_C_API_ON_TPU' not in os.environ:
+    os.environ['JAX_USE_PJRT_C_API_ON_TPU'] = 'true'
