@@ -22,23 +22,25 @@ import weakref
 import numpy as np
 
 import jax.numpy as jnp
-from jax import tree_util
 from jax import lax
+
 from jax._src import core
 from jax._src import effects
 from jax._src import linear_util as lu
 from jax._src import mesh as mesh_lib
-from jax._src import pjit
+from jax._src import sharding_impls
+from jax._src import tree_util
 from jax._src import util
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
+from jax._src.interpreters import partial_eval as pe
 from jax._src.lib import xla_client as xc
 from jax._src.lib.mlir import ir
 from jax._src.lib.mlir.dialects import hlo
 from jax._src.sharding import Sharding
-from jax._src.sharding_impls import GSPMDSharding, NamedSharding
-from jax._src.interpreters import partial_eval as pe
+from jax._src.sharding_impls import (GSPMDSharding, NamedSharding,
+                                     parse_flatten_op_sharding)
 
 # pytype: disable=import-error
 try:
@@ -122,13 +124,16 @@ ad.primitive_transposes[debug_callback_p] = debug_callback_transpose_rule
 def debug_callback_lowering(ctx, *args, effect, callback, **params):
 
   axis_context = ctx.module_context.axis_context
-  if (isinstance(axis_context, mlir.SPMDAxisContext) and
+  if (isinstance(axis_context, sharding_impls.SPMDAxisContext) and
         set(axis_context.manual_axes) == set(axis_context.mesh.axis_names)):
     # If we have fully manual sharding during lowering, that means the JAX
     # program has per-device semantics, so we run the callback on each device.
     sharding = xc.OpSharding()
     sharding.type = xc.OpSharding.Type.MANUAL
-  elif isinstance(axis_context, (mlir.ShardingContext, mlir.SPMDAxisContext)):
+  elif isinstance(
+      axis_context,
+      (sharding_impls.ShardingContext, sharding_impls.SPMDAxisContext),
+  ):
     # If we have fully automatic sharding during lowering, that means the JAX
     # program has bulk array semantics, so we run the callback with a MAXIMAL
     # sharding and hence execute it only once on the full logical value).
@@ -195,9 +200,11 @@ pe.partial_eval_jaxpr_custom_rules[debug_callback_p] = (
     _debug_callback_partial_eval_custom)
 
 def debug_callback(callback: Callable[..., Any], *args: Any,
-                   ordered: bool = False, **kwargs: Any):
+                   ordered: bool = False, **kwargs: Any) -> None:
   """Calls a stageable Python callback.
 
+  For more explanation, see `External Callbacks`_.
+
   `debug_callback` enables you to pass in a Python function that can be called
   inside of a staged JAX program. A `debug_callback` follows existing JAX
   transformation *pure* operational semantics, which are therefore unaware of
@@ -217,8 +224,16 @@ def debug_callback(callback: Callable[..., Any], *args: Any,
       staged out computation will enforce ordering of this callback w.r.t.
       other ordered callbacks.
     **kwargs: The keyword arguments to the callback.
+
   Returns:
-    The value of `callback(*args, **kwargs)`.
+    None
+
+  See Also:
+    - :func:`jax.experimental.io_callback`: callback designed for impure functions.
+    - :func:`jax.pure_callback`: callback designed for pure functions.
+    - :func:`jax.debug.print`: callback designed for printing.
+
+  .. _External Callbacks: https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html
   """
   flat_args, in_tree = tree_util.tree_flatten((args, kwargs))
   effect = ordered_debug_effect if ordered else debug_effect
@@ -226,8 +241,7 @@ def debug_callback(callback: Callable[..., Any], *args: Any,
     args, kwargs = tree_util.tree_unflatten(in_tree, flat_args)
     callback(*args, **kwargs)
     return []
-  return debug_callback_p.bind(*flat_args, callback=_flat_callback,
-                               effect=effect)
+  debug_callback_p.bind(*flat_args, callback=_flat_callback, effect=effect)
 
 class _DebugPrintFormatChecker(string.Formatter):
 
@@ -311,44 +325,31 @@ def _inspect_sharding_lowering_rule(ctx: mlir.LoweringRuleContext, value, *,
   mesh = mesh_lib.thread_resources.env.physical_mesh
   axis_context = ctx.module_context.axis_context
 
-  if isinstance(axis_context, mlir.ShardingContext):
+  if isinstance(axis_context, sharding_impls.ShardingContext):
     devices = axis_context.device_assignment
-  elif isinstance(axis_context, mlir.SPMDAxisContext):
+  elif isinstance(axis_context, sharding_impls.SPMDAxisContext):
     devices = list(axis_context.mesh.devices.flat)
   else:
     raise NotImplementedError(type(axis_context))
 
-  def _op_sharding_callback(op_sharding: xc.OpSharding):
+  # If we have a nontrivial parallel computation, we need to wait until the SPMD
+  # partitioner calls back with the `HloSharding.
+  def _hlo_sharding_callback(hlo_sharding: xc.HloSharding):
     if mesh.empty:
-      return callback(GSPMDSharding(
-        devices, op_sharding))
-    pspec = pjit.parse_flatten_op_sharding(
-        op_sharding, mesh)[0].get_partition_spec()
+      return callback(GSPMDSharding(devices, hlo_sharding))
+    pspec = parse_flatten_op_sharding(hlo_sharding, mesh)[0].get_partition_spec()
     return callback(NamedSharding(mesh, pspec))
 
   if len(devices) == 1:
-    # If we only have one device in our computation, we can construct a trivial
-    # OpSharding and call it right now.
-    trivial_sharding = xc.OpSharding()
-    trivial_sharding.type = xc.OpSharding.Type.REPLICATED
-    _op_sharding_callback(trivial_sharding)
+    # If we only have one device in our computation, we can construct a
+    # replicated HloSharding and call it right now.
+    _hlo_sharding_callback(sharding_impls.get_replicated_hlo_sharding())
     return []
 
-  # If we have a nontrivial parallel computation, we need to wait until the SPMD
-  # partitioner calls back with the `HloSharding.
-  def _hlo_sharding_callback(hlo_sharding):
-    op_sharding = hlo_sharding.to_proto()
-    return _op_sharding_callback(op_sharding)
-
-  # Here we store information in a container that we store globally so the
-  # custom partitioning code can access it.
-  sharding_callback_info = ShardingCallbackInfo(_hlo_sharding_callback,
-                                                ctx.module_context)
-  key = str(id(sharding_callback_info))
-  sharding_callbacks[key] = sharding_callback_info
-  # We need to make sure `sharding_callback_info` is still alive when the SPMD
-  # partitioner runs so we keep it alive by attaching it to the executable.
-  ctx.module_context.add_keepalive(sharding_callback_info)
+  key = xc.encode_inspect_sharding_callback(_hlo_sharding_callback)
+  # We need to make sure `_hlo_sharding_callback` is still alive when the SPMD
+  # partitioner runs so we keep it alive by attaching it to the executable.    #
+  ctx.module_context.add_keepalive(_hlo_sharding_callback)
 
   hlo.CustomCallOp([value.type], [value],
                    call_target_name=ir.StringAttr.get(
@@ -399,11 +400,6 @@ def inspect_sharding_infer_sharding_from_operands(arg_shapes, arg_shardings,
   del arg_shapes, shape, backend_string
   return arg_shardings[0]
 
-xc.register_custom_call_partitioner(  # pytype: disable=module-attr
-    _INSPECT_SHARDING_CALL_NAME, inspect_sharding_prop_user_sharding,
-    inspect_sharding_partition, inspect_sharding_infer_sharding_from_operands,
-    True)
-
 def _slice_to_chunk_idx(size: int, slc: slice) -> int:
   if slc.stop == slc.start == None:
     return 0
