@@ -14,6 +14,10 @@
 from __future__ import annotations
 
 from jax._src import core
+from jax._src.util import set_module
+
+export = set_module('jax.errors')
+
 
 class _JAXErrorMixin:
   """Mixin for JAX-specific errors"""
@@ -29,14 +33,17 @@ class _JAXErrorMixin:
     super().__init__(error_msg)  # type: ignore
 
 
+@export
 class JAXTypeError(_JAXErrorMixin, TypeError):
   pass
 
 
+@export
 class JAXIndexError(_JAXErrorMixin, IndexError):
   pass
 
 
+@export
 class ConcretizationTypeError(JAXTypeError):
   """
   This error occurs when a JAX Tracer object is used in a context where a
@@ -180,6 +187,7 @@ class ConcretizationTypeError(JAXTypeError):
         f"{tracer}\n{context}{tracer._origin_msg()}\n")
 
 
+@export
 class NonConcreteBooleanIndexError(JAXIndexError):
   """
   This error occurs when a program attempts to use non-concrete boolean indices
@@ -271,6 +279,7 @@ class NonConcreteBooleanIndexError(JAXIndexError):
         f"Array boolean indices must be concrete; got {tracer}\n")
 
 
+@export
 class TracerArrayConversionError(JAXTypeError):
   """
   This error occurs when a program attempts to convert a JAX Tracer object into
@@ -351,6 +360,7 @@ class TracerArrayConversionError(JAXTypeError):
         f"the JAX Tracer object {tracer}{tracer._origin_msg()}")
 
 
+@export
 class TracerIntegerConversionError(JAXTypeError):
   """
   This error can occur when a JAX Tracer object is used in a context where a
@@ -443,6 +453,7 @@ class TracerIntegerConversionError(JAXTypeError):
         f"The __index__() method was called on the JAX Tracer object {tracer}")
 
 
+@export
 class UnexpectedTracerError(JAXTypeError):
   """
   This error occurs when you use a JAX value that has leaked out of a function.
