@@ -22,19 +22,19 @@ import itertools as it
 import math
 import operator
 import re
-from typing import (Any, Callable, Dict, NamedTuple, Optional, Protocol,
+from typing import (Any, Callable, Dict, Optional, Protocol,
                     Sequence, Set, Type, Tuple, Union)
 
 import numpy as np
 
-from jax.config import config
+from jax._src.config import config
 
 from jax._src import core
-from jax._src import device_array
 from jax._src import dtypes
 from jax._src import source_info_util
 from jax._src.abstract_arrays import numpy_scalar_types
 from jax._src.core import ConcreteArray, ShapedArray
+from jax._src.sharding_impls import AxisEnv
 from jax._src.util import safe_zip, safe_map
 
 from jax._src.typing import Shape
@@ -54,11 +54,10 @@ def identity(x): return x
 
 _scalar_types = dtypes.python_scalar_dtypes.keys()
 
-def _make_array_shape(a: ShapedArray) -> Sequence[xc.Shape]:
-  if a.dtype == dtypes.float0:
-    return (xc.Shape.array_shape(np.dtype('bool'), a.shape),)
-  else:
-    return (xc.Shape.array_shape(a.dtype, a.shape),)
+def _make_array_shape(aval: ShapedArray) -> Sequence[xc.Shape]:
+  aval = core.physical_aval(aval)
+  dtype = np.dtype('bool') if aval.dtype == dtypes.float0 else aval.dtype
+  return (xc.Shape.array_shape(dtype, aval.shape),)
 
 def get_canonical_source_file(frame: source_info_util.Frame):
   source_file = frame.file_name
@@ -111,7 +110,7 @@ def sharding_to_proto(sharding: SpatialSharding):
   else:
     proto.type = xc.OpSharding.Type.OTHER
     proto.tile_assignment_dimensions = list(sharding)  # type: ignore
-    proto.tile_assignment_devices = list(range(np.product(sharding)))  # type: ignore
+    proto.tile_assignment_devices = list(range(np.prod(sharding)))  # type: ignore
   return proto
 
 def tuple_sharding_proto(elems):
@@ -180,8 +179,6 @@ def _canonicalize_python_scalar_dtype(typ, x):
       x, dtypes.canonicalize_dtype(dtypes._scalar_type_to_dtype(typ, x)))
 
 canonicalize_dtype_handlers: Dict[Any, Callable] = {}
-for t in device_array.device_array_types:
-  canonicalize_dtype_handlers[t] = identity
 canonicalize_dtype_handlers.update(
     (t, _canonicalize_ndarray_dtype) for t in numpy_scalar_types)
 canonicalize_dtype_handlers[np.ndarray] = _canonicalize_ndarray_dtype
@@ -220,8 +217,6 @@ def _make_shaped_array_for_numpy_array(x: np.ndarray) -> ShapedArray:
 
 
 pytype_aval_mappings: Dict[Any, Callable[[Any], core.AbstractValue]] = {}
-for t in device_array.device_array_types:
-  pytype_aval_mappings[t] = operator.attrgetter('aval')
 pytype_aval_mappings[core.DArray] = operator.attrgetter('_aval')
 pytype_aval_mappings[core.Token] = lambda _: core.abstract_token
 pytype_aval_mappings.update((t, _make_shaped_array_for_numpy_scalar)
@@ -259,13 +254,6 @@ def primitive_subcomputation(platform: str, axis_env: 'AxisEnv',
 
 ### compiling jaxprs
 
-
-class AxisEnv(NamedTuple):
-  """Represents a pmap mesh (only along the replica axes)."""
-  nreps: int
-  names: Tuple[Any, ...]
-  sizes: Tuple[int, ...]
-
 @dataclasses.dataclass
 class TranslationContext:
   builder: xc.XlaBuilder
@@ -277,8 +265,6 @@ class TranslationContext:
 
   def replace(self, **kw): return dataclasses.replace(self, **kw)
 
-
-
 def xla_destructure(c, ans):
   num_elements = len(c.get_shape(ans).tuple_shapes())
   return [xops.GetTupleElement(ans, i) for i in range(num_elements)]
@@ -344,9 +330,6 @@ def jaxpr_collectives(jaxpr):
 
 ### xla_call underlying jit
 
-# TODO(yashkatariya): Remove after 1 month from March 23, 2023.
-xla_call_p: core.CallPrimitive = core.CallPrimitive('xla_call')
-
 
 def xla_call_partial_eval_update_params(
     params: core.ParamDict, kept_inputs: Sequence[bool], num_new_inputs: int
@@ -459,9 +442,3 @@ class _BackendSpecificTranslationsAdapter(defaultdict):
 
 backend_specific_translations: Dict[str, _TranslationRuleAdapter]
 backend_specific_translations = _BackendSpecificTranslationsAdapter()
-
-
-# TODO(yashkatariya): Delete this.
-def device_put(x, device=None):
-  from jax._src import api
-  return api.device_put(x, device)
