@@ -22,15 +22,11 @@ import jax
 import weakref
 from jax._src import core
 from jax._src import linear_util as lu
-from jax.config import config
+from jax import config  # type: ignore[no-redef]
 from jax._src.core import ConcreteArray, ShapedArray, raise_to_shaped
-from jax.interpreters import ad
-from jax.interpreters import batching
-from jax._src.interpreters import mlir
-from jax._src.interpreters import partial_eval as pe
-from jax._src.interpreters import xla
 from jax.tree_util import (tree_flatten, tree_unflatten, treedef_is_leaf,
                            tree_map, tree_flatten_with_path, keystr)
+from jax._src.api_util import shaped_abstractify
 from jax._src.tree_util import equality_errors
 from jax._src import ad_checkpoint
 from jax._src import ad_util
@@ -40,6 +36,11 @@ from jax._src import dtypes
 from jax._src import effects
 from jax._src import source_info_util
 from jax._src import util
+from jax._src.interpreters import ad
+from jax._src.interpreters import batching
+from jax._src.interpreters import mlir
+from jax._src.interpreters import partial_eval as pe
+from jax._src.interpreters import xla
 from jax._src.lax import lax
 from jax._src.lax import slicing
 from jax._src.lax import windowed_reductions
@@ -692,7 +693,9 @@ def _scan_partial_eval(trace, *tracers, reverse, length, num_consts, num_carry,
 
 def _maybe_put(x):
   if isinstance(x, np.ndarray):
-    return jax.device_put(x, jax.devices('cpu')[0])
+    return dispatch._put_x(
+        x, jax.sharding.SingleDeviceSharding(jax.devices('cpu')[0]),
+        shaped_abstractify(x), False)
   else:
     return x
 
@@ -843,10 +846,14 @@ def _scan_dce_rule(used_outputs: List[bool], eqn: core.JaxprEqn
                     num_carry=sum(used_carry_in), linear=tuple(new_linear),
                     jaxpr=core.ClosedJaxpr(jaxpr_dce, jaxpr.consts))
   # TODO(mattjj,sharadmv): don't assume effects are never DCE'd?
+  new_invars = [v for v, used in zip(eqn.invars, used_inputs) if used]
+  new_outvars = [v for v, used in zip(eqn.outvars, used_outputs) if used]
+  _, new_effects = eqn.primitive.abstract_eval(*[v.aval for v in new_invars],
+                                               **new_params)
   new_eqn = pe.new_jaxpr_eqn(
-      [v for v, used in zip(eqn.invars, used_inputs) if used],
-      [v for v, used in zip(eqn.outvars, used_outputs) if used],
-      eqn.primitive, new_params, eqn.effects, eqn.source_info)
+      new_invars,
+      new_outvars,
+      eqn.primitive, new_params, new_effects, eqn.source_info)
   assert len(new_eqn.invars ) == len(new_params['jaxpr'].in_avals )
   assert len(new_eqn.outvars) == len(new_params['jaxpr'].out_avals)
   return used_inputs, new_eqn
@@ -1464,8 +1471,8 @@ def _while_partial_eval_custom(saveable, unks_in, inst_in, eqn):
 
 def _while_transpose_error(*_, **kwargs):
   raise ValueError("Reverse-mode differentiation does not work for "
-                   "lax.while_loop or lax.fori_loop. "
-                   "Try using lax.scan instead.")
+                   "lax.while_loop or lax.fori_loop with dynamic start/stop values. "
+                   "Try using lax.scan, or using fori_loop with static start/stop.")
 
 # For a while loop with ordered effects in the cond, we need a special
 # lowering. Fundamentally, we'd like to rewrite a while loop that looks like
@@ -1640,12 +1647,10 @@ def _pred_bcast_select_hlo(ctx,
     assert x.type == y.type, (x.type, y.type)
     assert (pred_aval.shape == x_y_aval.shape[:len(pred_aval.shape)]), (
             pred_aval.shape, x_y_aval)
-    if core.is_opaque_dtype(x_y_aval.dtype):
-      x_y_shape = mlir.aval_to_ir_type(x_y_aval).shape
-    else:
-      x_y_shape = x_y_aval.shape
-    bcast_pred = mlir.broadcast_in_dim(ctx, pred, core.DShapedArray(x_y_shape, np.dtype(np.bool_)),
-                                       broadcast_dimensions=list(range(len(pred_aval.shape))))
+    x_y_aval = core.physical_aval(x_y_aval)
+    bcast_pred = mlir.broadcast_in_dim(
+        ctx, pred, core.DShapedArray(x_y_aval.shape, np.dtype(np.bool_)),
+        broadcast_dimensions=list(range(len(pred_aval.shape))))
     return hlo.SelectOp(bcast_pred, x, y).results
 
 ### fori_loop
@@ -1724,13 +1729,29 @@ def fori_loop(lower, upper, body_fun, init_val):
   """
   if not callable(body_fun):
     raise TypeError("lax.fori_loop: body_fun argument should be callable.")
+
   # TODO(phawkins): perhaps do more type checking here, better error messages.
   lower_dtype = dtypes.canonicalize_dtype(lax.dtype(lower))
   upper_dtype = dtypes.canonicalize_dtype(lax.dtype(upper))
-  if lower_dtype != upper_dtype:
-    msg = ("lower and upper arguments to fori_loop must have equal types, "
-           "got {} and {}")
-    raise TypeError(msg.format(lower_dtype.name, upper_dtype.name))
+  if lower_dtype == upper_dtype:
+    dtype = lower_dtype
+  else:
+    # As a special case: allow promotion of weak integers (e.g., Python scalars)
+    # This improves the ergonomics if one but not both of the loop bounds is a
+    # scalar.
+    dtype = None
+    if (np.issubdtype(lower_dtype, np.signedinteger) and
+        np.issubdtype(upper_dtype, np.signedinteger)):
+      lower_weak = dtypes.is_weakly_typed(lower)
+      upper_weak = dtypes.is_weakly_typed(upper)
+      if lower_weak and not upper_weak:
+        dtype = upper_dtype
+      elif not lower_weak and upper_weak:
+        dtype = lower_dtype
+
+    if dtype is None:
+      raise TypeError("lower and upper arguments to fori_loop must have equal "
+                      f"types, got {lower_dtype.name} and {upper_dtype.name}")
 
   # If we can specialize on the trip count, call scan instead of a while_loop
   # to enable efficient reverse-mode differentiation.
@@ -1753,9 +1774,14 @@ def fori_loop(lower, upper, body_fun, init_val):
 
     (_, result), _ = scan(_fori_scan_body_fun(body_fun), (lower_, init_val),
                           None, length=upper_ - lower_)
-  else:
-    _, _, result = while_loop(_fori_cond_fun, _fori_body_fun(body_fun),
-                              (lower, upper, init_val))
+    return result
+
+  if lower_dtype != dtype:
+    lower = lax.convert_element_type(lower, dtype)
+  if upper_dtype != dtype:
+    upper = lax.convert_element_type(upper, dtype)
+  _, _, result = while_loop(_fori_cond_fun, _fori_body_fun(body_fun),
+                            (lower, upper, init_val))
   return result
 
 ### map and miscellaneous rules
