@@ -20,17 +20,17 @@ from typing import Union, Sequence
 import numpy as np
 
 from jax import lax
-from jax.interpreters import mlir
-from jax.interpreters import xla
 
+from jax._src import dispatch
 from jax._src.api import jit, linear_transpose, ShapeDtypeStruct
 from jax._src.core import Primitive, is_constant_shape
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
+from jax._src.interpreters import mlir
 from jax._src.lib.mlir.dialects import hlo
 from jax._src.lib import xla_client
 from jax._src.lib import ducc_fft
-from jax._src.numpy.util import _promote_dtypes_complex, _promote_dtypes_inexact
+from jax._src.numpy.util import promote_dtypes_complex, promote_dtypes_inexact
 
 __all__ = [
   "fft",
@@ -61,9 +61,9 @@ def fft(x, fft_type: Union[xla_client.FftType, str], fft_lengths: Sequence[int])
   if typ == xla_client.FftType.RFFT:
     if np.iscomplexobj(x):
       raise ValueError("only real valued inputs supported for rfft")
-    x, = _promote_dtypes_inexact(x)
+    x, = promote_dtypes_inexact(x)
   else:
-    x, = _promote_dtypes_complex(x)
+    x, = promote_dtypes_complex(x)
   if len(fft_lengths) == 0:
     # XLA FFT doesn't support 0-rank.
     return x
@@ -71,11 +71,10 @@ def fft(x, fft_type: Union[xla_client.FftType, str], fft_lengths: Sequence[int])
   return fft_p.bind(x, fft_type=typ, fft_lengths=fft_lengths)
 
 def _fft_impl(x, fft_type, fft_lengths):
-  return xla.apply_primitive(fft_p, x, fft_type=fft_type, fft_lengths=fft_lengths)
+  return dispatch.apply_primitive(fft_p, x, fft_type=fft_type, fft_lengths=fft_lengths)
 
 _complex_dtype = lambda dtype: (np.zeros((), dtype) + np.zeros((), np.complex64)).dtype
 _real_dtype = lambda dtype: np.finfo(dtype).dtype
-_is_even = lambda x: x % 2 == 0
 
 def fft_abstract_eval(x, fft_type, fft_lengths):
   if len(fft_lengths) > x.ndim:
