@@ -16,26 +16,26 @@
 from functools import partial
 import math
 from operator import index
-from typing import Optional, Sequence, Union
+from typing import Optional, Sequence, Tuple, Union
 import warnings
 
 import numpy as np
 
-import jax
 import jax.numpy as jnp
 from jax import lax
-from jax.config import config
-from jax.interpreters import mlir
 from jax.numpy.linalg import cholesky, svd, eigh
 
+from jax._src import config as config_lib
 from jax._src import core
 from jax._src import dtypes
 from jax._src import prng
 from jax._src import xla_bridge
 from jax._src.api import jit, vmap
+from jax._src.config import config
 from jax._src.core import NamedShape
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
+from jax._src.interpreters import mlir
 from jax._src.lax import lax as lax_internal
 from jax._src.numpy.lax_numpy import _convert_and_clip_integer
 from jax._src.numpy.util import _arraylike, check_arraylike, promote_dtypes_inexact
@@ -48,11 +48,13 @@ IntegerArray = ArrayLike
 # TODO: Import or define these to match
 # https://github.com/numpy/numpy/blob/main/numpy/typing/_dtype_like.py.
 DTypeLikeInt = DTypeLike
+DTypeLikeUInt = DTypeLike
 DTypeLikeFloat = DTypeLike
 Shape = Sequence[int]
 
 # TODO(frostig): simplify once we always enable_custom_prng
 KeyArray = Union[Array, prng.PRNGKeyArray]
+PRNGKeyArray = prng.PRNGKeyArray
 
 UINT_DTYPES = prng.UINT_DTYPES
 
@@ -65,7 +67,7 @@ def _isnan(x: ArrayLike) -> Array:
   return lax.ne(x, x)
 
 
-def _check_prng_key(key):
+def _check_prng_key(key) -> Tuple[prng.PRNGKeyArray, bool]:
   # TODO(frostig): remove once we always enable_custom_prng
   if isinstance(key, prng.PRNGKeyArray):
     return key, False
@@ -114,6 +116,22 @@ def default_prng_impl():
 
 ### key operations
 
+def key(seed: Union[int, Array]) -> PRNGKeyArray:
+  """Create a pseudo-random number generator (PRNG) key given an integer seed.
+
+  The result is a scalar array with a key that indicates the default PRNG
+  implementation, as determined by the ``jax_default_prng_impl`` config flag.
+
+  Args:
+    seed: a 64- or 32-bit integer used as the value of the key.
+
+  Returns:
+    A scalar PRNG key array, consumable by random functions as well as ``split``
+    and ``fold_in``.
+  """
+  # TODO(frostig): Take impl as optional argument
+  impl = default_prng_impl()
+  return prng.seed_with_impl(impl, seed)
 
 def PRNGKey(seed: Union[int, Array]) -> KeyArray:
   """Create a pseudo-random number generator (PRNG) key given an integer seed.
@@ -127,11 +145,12 @@ def PRNGKey(seed: Union[int, Array]) -> KeyArray:
   Returns:
     A PRNG key, consumable by random functions as well as ``split``
     and ``fold_in``.
-
   """
   impl = default_prng_impl()
+  if isinstance(seed, prng.PRNGKeyArray):
+    raise TypeError("PRNGKey accepts a scalar seed, but was given a PRNGKeyArray.")
   if np.ndim(seed):
-    raise TypeError("PRNGKey accepts a scalar seed, but was given an array of"
+    raise TypeError("PRNGKey accepts a scalar seed, but was given an array of "
                     f"shape {np.shape(seed)} != (). Use jax.vmap for batching")
   key = prng.seed_with_impl(impl, seed)
   return _return_prng_keys(True, key)
@@ -240,6 +259,35 @@ def _check_shape(name: str, shape: Union[Shape, NamedShape], *param_shapes) -> N
       raise ValueError(msg.format(name, shape_, shape))
 
 
+def bits(key: KeyArray,
+         shape: Shape = (),
+         dtype: Optional[DTypeLikeUInt] = None) -> Array:
+  """Sample uniform bits in the form of unsigned integers.
+
+  Args:
+    key: a PRNG key used as the random key.
+    shape: optional, a tuple of nonnegative integers representing the result
+      shape. Default ``()``.
+    dtype: optional, an unsigned integer dtype for the returned values (default
+      ``uint64`` if ``jax_enable_x64`` is true, otherwise ``uint32``).
+
+  Returns:
+    A random array with the specified shape and dtype.
+  """
+  key, _ = _check_prng_key(key)
+  if dtype is None:
+    dtype = dtypes.canonicalize_dtype(jnp.uint)
+  else:
+    dtypes.check_user_dtype_supported(dtype)
+  if not dtypes.issubdtype(dtype, np.unsignedinteger):
+    raise ValueError("dtype argument to `bits` must be an unsigned int dtype, "
+                     f"got {dtype}")
+  dtype = dtypes.canonicalize_dtype(dtype)
+  shape = core.canonicalize_shape(shape)
+  bit_width = dtype.itemsize * 8
+  return _random_bits(key, bit_width, shape)
+
+
 def uniform(key: KeyArray,
             shape: Union[Shape, NamedShape] = (),
             dtype: DTypeLikeFloat = dtypes.float_,
@@ -267,7 +315,7 @@ def uniform(key: KeyArray,
   shape = core.as_named_shape(shape)
   return _uniform(key, shape, dtype, minval, maxval)  # type: ignore
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _uniform(key, shape, dtype, minval, maxval) -> Array:
   _check_shape("uniform", shape)
   if not jnp.issubdtype(dtype, np.floating):
@@ -331,7 +379,7 @@ def randint(key: KeyArray,
   shape = core.canonicalize_shape(shape)
   return _randint(key, shape, minval, maxval, dtype)
 
-@partial(jit, static_argnums=(1, 4), inline=True)
+@partial(jit, static_argnums=(1, 4))
 def _randint(key, shape, minval, maxval, dtype) -> Array:
   _check_shape("randint", shape, np.shape(minval), np.shape(maxval))
   if not jnp.issubdtype(dtype, np.integer):
@@ -443,7 +491,7 @@ def permutation(key: KeyArray,
   return jnp.take(x, ind, axis, unique_indices=True)
 
 
-@partial(jit, static_argnums=(2,), inline=True)
+@partial(jit, static_argnums=(2,))
 def _shuffle(key, x, axis) -> Array:
   # On parallel architectures, Fisher-Yates is more expensive than doing
   # multiple sorts. This algorithm is based on one developed and analyzed by
@@ -537,7 +585,7 @@ def choice(key: KeyArray,
     if replace:
       p_cuml = jnp.cumsum(p_arr)
       r = p_cuml[-1] * (1 - uniform(key, shape, dtype=p_cuml.dtype))
-      ind = jnp.searchsorted(p_cuml, r)
+      ind = jnp.searchsorted(p_cuml, r).astype(int)
     else:
       # Gumbel top-k trick: https://timvieira.github.io/blog/post/2019/09/16/algorithms-for-sampling-without-replacement/
       g = -gumbel(key, (n_inputs,), dtype=p_arr.dtype) - jnp.log(p_arr)
@@ -578,7 +626,7 @@ def normal(key: KeyArray,
   shape = core.as_named_shape(shape)
   return _normal(key, shape, dtype)  # type: ignore
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _normal(key, shape, dtype) -> Array:
   if dtypes.issubdtype(dtype, np.complexfloating):
     sqrt2 = np.array(np.sqrt(2), dtype)
@@ -591,7 +639,7 @@ def _normal(key, shape, dtype) -> Array:
   else:
     return _normal_real(key, shape, dtype) # type: ignore
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _normal_real(key, shape, dtype) -> Array:
   _check_shape("normal", shape)
   lo = np.nextafter(np.array(-1., dtype), np.array(0., dtype), dtype=dtype)
@@ -649,7 +697,7 @@ def multivariate_normal(key: KeyArray,
     shape = core.canonicalize_shape(shape)
   return _multivariate_normal(key, mean, cov, shape, dtype, method)  # type: ignore
 
-@partial(jit, static_argnums=(3, 4, 5), inline=True)
+@partial(jit, static_argnums=(3, 4, 5))
 def _multivariate_normal(key, mean, cov, shape, dtype, method) -> Array:
   if not np.ndim(mean) >= 1:
     msg = "multivariate_normal requires mean.ndim >= 1, got mean.ndim == {}"
@@ -677,7 +725,7 @@ def _multivariate_normal(key, mean, cov, shape, dtype, method) -> Array:
   else: # 'cholesky'
     factor = cholesky(cov)
   normal_samples = normal(key, shape + mean.shape[-1:], dtype)
-  with jax.numpy_rank_promotion('allow'):
+  with config_lib.numpy_rank_promotion('allow'):
     result = mean + jnp.einsum('...ij,...j->...i', factor, normal_samples)
   return result
 
@@ -723,7 +771,7 @@ def truncated_normal(key: KeyArray,
     shape = core.as_named_shape(shape)
   return _truncated_normal(key, lower, upper, shape, dtype)  # type: ignore
 
-@partial(jit, static_argnums=(3, 4), inline=True)
+@partial(jit, static_argnums=(3, 4))
 def _truncated_normal(key, lower, upper, shape, dtype) -> Array:
   if shape is None:
     shape = lax.broadcast_shapes(np.shape(lower), np.shape(upper))
@@ -781,7 +829,7 @@ def bernoulli(key: KeyArray,
   p = lax.convert_element_type(p, dtype)
   return _bernoulli(key, p, shape)  # type: ignore
 
-@partial(jit, static_argnums=(2,), inline=True)
+@partial(jit, static_argnums=(2,))
 def _bernoulli(key, p, shape) -> Array:
   if shape is None:
     # TODO: Use the named part of `p` as well
@@ -882,7 +930,7 @@ def cauchy(key: KeyArray,
   shape = core.canonicalize_shape(shape)
   return _cauchy(key, shape, dtype)
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _cauchy(key, shape, dtype) -> Array:
   _check_shape("cauchy", shape)
   u = uniform(key, shape, dtype, minval=jnp.finfo(dtype).eps, maxval=1.)
@@ -934,7 +982,7 @@ def dirichlet(key: KeyArray,
     shape = core.canonicalize_shape(shape)
   return _dirichlet(key, alpha, shape, dtype)
 
-@partial(jit, static_argnums=(2, 3), inline=True)
+@partial(jit, static_argnums=(2, 3))
 def _dirichlet(key, alpha, shape, dtype) -> Array:
   if not np.ndim(alpha) >= 1:
     msg = "dirichlet requires alpha.ndim >= 1, got alpha.ndim == {}"
@@ -991,7 +1039,7 @@ def exponential(key: KeyArray,
   shape = core.canonicalize_shape(shape)
   return _exponential(key, shape, dtype)
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _exponential(key, shape, dtype) -> Array:
   _check_shape("exponential", shape)
   u = uniform(key, shape, dtype)
@@ -1218,7 +1266,7 @@ def loggamma(key: KeyArray,
   return _gamma(key, a, shape=shape, dtype=dtype, log_space=True)
 
 
-@partial(jit, static_argnames=('shape', 'dtype', 'log_space'), inline=True)
+@partial(jit, static_argnames=('shape', 'dtype', 'log_space'))
 def _gamma(key, a, shape, dtype, log_space=False) -> Array:
   if shape is None:
     shape = np.shape(a)
@@ -1231,7 +1279,7 @@ def _gamma(key, a, shape, dtype, log_space=False) -> Array:
   return random_gamma_p.bind(key, a, log_space=log_space)
 
 
-@partial(jit, static_argnums=(2, 3, 4), inline=True)
+@partial(jit, static_argnums=(2, 3, 4))
 def _poisson_knuth(key, lam, shape, dtype, max_iters) -> Array:
   # Knuth's algorithm for generating Poisson random variates.
   # Reference:
@@ -1254,7 +1302,7 @@ def _poisson_knuth(key, lam, shape, dtype, max_iters) -> Array:
   return (k - 1).astype(dtype)
 
 
-@partial(jit, static_argnums=(2, 3, 4), inline=True)
+@partial(jit, static_argnums=(2, 3, 4))
 def _poisson_rejection(key, lam, shape, dtype, max_iters) -> Array:
   # Transformed rejection due to Hormann.
   # Reference:
@@ -1297,7 +1345,7 @@ def _poisson_rejection(key, lam, shape, dtype, max_iters) -> Array:
   return k.astype(dtype)
 
 
-@partial(jit, static_argnums=(2, 3), inline=True)
+@partial(jit, static_argnums=(2, 3))
 def _poisson(key, lam, shape, dtype) -> Array:
   # The implementation matches TensorFlow and NumPy:
   # https://github.com/tensorflow/tensorflow/blob/v2.2.0-rc3/tensorflow/core/kernels/random_poisson_op.cc
@@ -1389,7 +1437,7 @@ def gumbel(key: KeyArray,
   shape = core.canonicalize_shape(shape)
   return _gumbel(key, shape, dtype)
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _gumbel(key, shape, dtype) -> Array:
   _check_shape("gumbel", shape)
   return -jnp.log(-jnp.log(
@@ -1466,7 +1514,7 @@ def laplace(key: KeyArray,
   shape = core.canonicalize_shape(shape)
   return _laplace(key, shape, dtype)
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _laplace(key, shape, dtype) -> Array:
   _check_shape("laplace", shape)
   u = uniform(
@@ -1502,7 +1550,7 @@ def logistic(key: KeyArray,
   shape = core.canonicalize_shape(shape)
   return _logistic(key, shape, dtype)
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _logistic(key, shape, dtype):
   _check_shape("logistic", shape)
   x = uniform(key, shape, dtype, minval=jnp.finfo(dtype).eps, maxval=1.)
@@ -1520,7 +1568,7 @@ def pareto(key: KeyArray,
   .. math::
      f(x; b) = b / x^{b + 1}
 
-  on the domain :math:`0 \le x < \infty` with :math:`b > 0`
+  on the domain :math:`1 \le x < \infty` with :math:`b > 0`
 
   Args:
     key: a PRNG key used as the random key.
@@ -1545,7 +1593,7 @@ def pareto(key: KeyArray,
     shape = core.canonicalize_shape(shape)
   return _pareto(key, b, shape, dtype)
 
-@partial(jit, static_argnums=(2, 3), inline=True)
+@partial(jit, static_argnums=(2, 3))
 def _pareto(key, b, shape, dtype) -> Array:
   if shape is None:
     shape = np.shape(b)
@@ -1592,7 +1640,7 @@ def t(key: KeyArray,
   shape = core.canonicalize_shape(shape)
   return _t(key, df, shape, dtype)
 
-@partial(jit, static_argnums=(2, 3), inline=True)
+@partial(jit, static_argnums=(2, 3))
 def _t(key, df, shape, dtype) -> Array:
   if shape is None:
     shape = np.shape(df)
@@ -1645,7 +1693,7 @@ def chisquare(key: KeyArray,
     shape = core.canonicalize_shape(shape)
   return _chisquare(key, df, shape, dtype)
 
-@partial(jit, static_argnums=(2, 3), inline=True)
+@partial(jit, static_argnums=(2, 3))
 def _chisquare(key, df, shape, dtype) -> Array:
   if shape is None:
     shape = np.shape(df)
@@ -1702,7 +1750,7 @@ def f(key: KeyArray,
     shape = core.canonicalize_shape(shape)
   return _f(key, dfnum, dfden, shape, dtype)
 
-@partial(jit, static_argnums=(3, 4), inline=True)
+@partial(jit, static_argnums=(3, 4))
 def _f(key, dfnum, dfden, shape, dtype) -> Array:
   if shape is None:
     shape =  lax.broadcast_shapes(np.shape(dfden), np.shape(dfnum))
@@ -1750,7 +1798,7 @@ def rademacher(key: KeyArray,
   return _rademacher(key, shape, dtype)
 
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _rademacher(key, shape, dtype) -> Array:
   bernoulli_samples = bernoulli(key=key, p=0.5, shape=shape).astype(dtype)
   return (2 * bernoulli_samples - 1).astype(dtype)
@@ -1788,7 +1836,7 @@ def maxwell(key: KeyArray,
   return _maxwell(key, shape, dtype)
 
 
-@partial(jit, static_argnums=(1, 2), inline=True)
+@partial(jit, static_argnums=(1, 2))
 def _maxwell(key, shape, dtype) -> Array:
   shape = shape + (3,)
   norm_rvs = normal(key=key, shape=shape, dtype=dtype)
@@ -1830,7 +1878,7 @@ def double_sided_maxwell(key: KeyArray,
   return _double_sided_maxwell(key, loc, scale, shape, dtype)
 
 
-@partial(jit, static_argnums=(3, 4), inline=True)
+@partial(jit, static_argnums=(3, 4))
 def _double_sided_maxwell(key, loc, scale, shape, dtype) -> Array:
   params_shapes = lax.broadcast_shapes(np.shape(loc), np.shape(scale))
   if not shape:
@@ -1881,7 +1929,7 @@ def weibull_min(key: KeyArray,
   return _weibull_min(key, scale, concentration, shape, dtype)
 
 
-@partial(jit, static_argnums=(3, 4), inline=True)
+@partial(jit, static_argnums=(3, 4))
 def _weibull_min(key, scale, concentration, shape, dtype) -> Array:
   random_uniform = uniform(
       key=key, shape=shape, minval=0, maxval=1, dtype=dtype)
@@ -2028,7 +2076,7 @@ def rayleigh(key: KeyArray,
     shape = core.canonicalize_shape(shape)
   return _rayleigh(key, scale, shape, dtype)
 
-@partial(jit, static_argnums=(2, 3), inline=True)
+@partial(jit, static_argnums=(2, 3))
 def _rayleigh(key, scale, shape, dtype) -> Array:
   if shape is None:
     shape = np.shape(scale)
@@ -2036,6 +2084,7 @@ def _rayleigh(key, scale, shape, dtype) -> Array:
     _check_shape("rayleigh", shape, np.shape(scale))
   u = uniform(key, shape, dtype)
   scale = scale.astype(dtype)
+  scale = jnp.broadcast_to(scale, shape)
   log_u = lax.log(u)
   n_two = _lax_const(scale, -2)
   sqrt_u = lax.sqrt(lax.mul(log_u, n_two))
@@ -2080,7 +2129,7 @@ def wald(key: KeyArray,
     shape = core.canonicalize_shape(shape)
   return _wald(key, mean, shape, dtype)
 
-@partial(jit, static_argnums=(2, 3), inline=True)
+@partial(jit, static_argnums=(2, 3))
 def _wald(key, mean, shape, dtype) -> Array:
   if shape is None:
     shape =  np.shape(mean)
@@ -2098,3 +2147,54 @@ def _wald(key, mean, shape, dtype) -> Array:
   x = mean + mean_sq * y / 2 - mean / 2 * sqrt_term
   w = lax.select(lax.le(z,  mean / (mean + x)), x, mean_sq / x)
   return w
+
+def geometric(key: KeyArray,
+              p: RealArray,
+              shape: Optional[Shape] = None,
+              dtype: DTypeLikeInt = dtypes.int_) -> Array:
+  r"""Sample Geometric random values with given shape and float dtype.
+
+  The values are returned according to the probability mass function:
+
+  .. math::
+      f(k;p) = p(1-p)^{k-1}
+
+  on the domain :math:`0 < p < 1`.
+
+  Args:
+    key: a PRNG key used as the random key.
+    p: a float or array of floats broadcast-compatible with ``shape``
+      representing the the probability of success of an individual trial.
+    shape: optional, a tuple of nonnegative integers specifying the result
+      shape. Must be broadcast-compatible with ``p``. The default
+      (None) produces a result shape equal to ``np.shape(p)``.
+    dtype: optional, a int dtype for the returned values (default int64 if
+      jax_enable_x64 is true, otherwise int32).
+
+  Returns:
+    A random array with the specified dtype and with shape given by ``shape`` if
+    ``shape`` is not None, or else by ``p.shape``.
+  """
+  key, _ = _check_prng_key(key)
+  if not dtypes.issubdtype(dtype, np.integer):
+    raise ValueError("dtype argument to `geometric` must be an int "
+                     f"dtype, got {dtype}")
+  dtype = dtypes.canonicalize_dtype(dtype)
+  if shape is not None:
+    shape = core.canonicalize_shape(shape)
+  return _geometric(key, p, shape, dtype)
+
+@partial(jit, static_argnums=(2, 3))
+def _geometric(key, p, shape, dtype) -> Array:
+  if shape is None:
+    shape =  np.shape(p)
+  else:
+    _check_shape("geometric", shape, np.shape(p))
+  check_arraylike("geometric", p)
+  p, = promote_dtypes_inexact(p)
+  u = uniform(key, shape, p.dtype)
+  log_u = lax.log(u)
+  log_one_minus_p = lax.log1p(-p)
+  log_one_minus_p = jnp.broadcast_to(log_one_minus_p, shape)
+  g = lax.floor(lax.div(log_u, log_one_minus_p)) + 1
+  return g.astype(dtype)
