@@ -18,14 +18,14 @@ import scipy.stats as osp_stats
 from jax import lax
 import jax.numpy as jnp
 from jax._src.lax.lax import _const as _lax_const
-from jax._src.numpy.util import _wraps, _promote_args_inexact
+from jax._src.numpy.util import _wraps, promote_args_inexact
 from jax._src.typing import Array, ArrayLike
 from jax.scipy.special import xlogy, xlog1py
 
 
 @_wraps(osp_stats.bernoulli.logpmf, update_doc=False)
 def logpmf(k: ArrayLike, p: ArrayLike, loc: ArrayLike = 0) -> Array:
-  k, p, loc = _promote_args_inexact("bernoulli.logpmf", k, p, loc)
+  k, p, loc = promote_args_inexact("bernoulli.logpmf", k, p, loc)
   zero = _lax_const(k, 0)
   one = _lax_const(k, 1)
   x = lax.sub(k, loc)
@@ -39,7 +39,7 @@ def pmf(k: ArrayLike, p: ArrayLike, loc: ArrayLike = 0) -> Array:
 
 @_wraps(osp_stats.bernoulli.cdf, update_doc=False)
 def cdf(k: ArrayLike, p: ArrayLike) -> Array:
-  k, p = _promote_args_inexact('bernoulli.cdf', k, p)
+  k, p = promote_args_inexact('bernoulli.cdf', k, p)
   zero, one = _lax_const(k, 0), _lax_const(k, 1)
   conds = [
     jnp.isnan(k) | jnp.isnan(p) | (p < zero) | (p > one),
@@ -52,7 +52,7 @@ def cdf(k: ArrayLike, p: ArrayLike) -> Array:
 
 @_wraps(osp_stats.bernoulli.ppf, update_doc=False)
 def ppf(q: ArrayLike, p: ArrayLike) -> Array:
-  q, p = _promote_args_inexact('bernoulli.ppf', q, p)
+  q, p = promote_args_inexact('bernoulli.ppf', q, p)
   zero, one = _lax_const(q, 0), _lax_const(q, 1)
   return jnp.where(
     jnp.isnan(q) | jnp.isnan(p) | (p < zero) | (p > one) | (q < zero) | (q > one),
