@@ -33,16 +33,17 @@ from __future__ import annotations
 import warnings
 
 from dataclasses import dataclass
-from typing import Any, Dict, List, NamedTuple, Optional, Protocol, Sequence, Tuple
+from typing import (
+    Any, Dict, List, NamedTuple, Optional, Protocol, Sequence, Tuple, Union)
 
 import jax
-from jax import tree_util
-from jax.interpreters import mlir
 
 from jax._src import core
 from jax._src import source_info_util
 from jax._src import traceback_util
+from jax._src import tree_util
 from jax._src import util
+from jax._src.interpreters import mlir
 from jax._src.lib.mlir import ir
 from jax._src.lib import xla_client as xc
 
@@ -55,6 +56,8 @@ xla_extension = xc._xla
 map, unsafe_map = util.safe_map, map
 zip, unsafe_zip = util.safe_zip, zip
 
+CompilerOptions = Dict[str, Union[str, bool]]
+
 
 # -- Internal protocols
 
@@ -145,7 +148,8 @@ class Executable(Protocol):
 class Lowering(Protocol):
   """Protocol for lowerings, which a user-facing ``Lowered`` encapsulates."""
 
-  def compile(self) -> Executable:
+  def compile(
+      self, compiler_options: Optional[CompilerOptions] = None) -> Executable:
     """Compile and return a corresponding ``Executable``."""
     raise NotImplementedError
 
@@ -194,6 +198,7 @@ class Lowering(Protocol):
     # TODO(frostig): improve annotation (arbitrary pytree)
     raise NotImplementedError
 
+
 # -- Internal adapters from XLA-related objects to the above protocols
 
 class XlaExecutable(Executable):
@@ -282,7 +287,9 @@ class XlaLowering(Lowering):
 
   def hlo(self) -> xc.XlaComputation:
     """Return an HLO representation of this computation."""
-    raise NotImplementedError("must override")
+    return xla_extension.mlir.mlir_module_to_xla_computation(
+        mlir.module_to_string(self.stablehlo()),
+        use_tuple_args=self.compile_args["tuple_args"])
 
   def mhlo(self) -> ir.Module:
     """Return an MHLO representation of this computation."""
@@ -295,7 +302,8 @@ class XlaLowering(Lowering):
     """Return a StableHLO representation of this computation."""
     raise NotImplementedError("must override")
 
-  def compile(self) -> Executable:
+  def compile(
+      self, compiler_options: Optional[CompilerOptions] = None) -> Executable:
     raise NotImplementedError("must override")
 
   def as_text(self, dialect: Optional[str] = None) -> str:
@@ -391,9 +399,7 @@ class Compiled(Stage):
     self.out_tree = out_tree
     self._params = CompiledCallParams(self._executable, self._no_kwargs,
                                       self.in_tree, self.out_tree)
-    self._cpp_call = self._executable.create_cpp_call(self._no_kwargs,
-                                                      self.in_tree,
-                                                      self.out_tree)
+    self._call = None
 
   def compiler_ir(self):
     """Post-compilation IR.
@@ -526,11 +532,17 @@ class Compiled(Stage):
     return outs, out_flat, args_flat
 
   def __call__(self, *args, **kwargs):
-    if self._cpp_call is not None:
-      return self._cpp_call(*args, **kwargs)
-
-    outs, _, _ = Compiled.call(self._params, *args, **kwargs)
-    return outs
+    if self._call is None:
+      self._call = self._executable.create_cpp_call(self._no_kwargs,
+                                                        self.in_tree,
+                                                        self.out_tree)
+      if self._call is None:
+        params = self._params
+        def cpp_call_fallback(*args, **kwargs):
+          outs, _, _ = Compiled.call(params, *args, **kwargs)
+          return outs
+        self._call = cpp_call_fallback
+    return self._call(*args, **kwargs)
 
 
 class Lowered(Stage):
@@ -583,24 +595,16 @@ class Lowered(Stage):
         out_tree,
         no_kwargs=no_kwargs)
 
-  def compile(self) -> Compiled:
+  def compile(
+      self, compiler_options: Optional[CompilerOptions] = None) -> Compiled:
     """Compile, returning a corresponding ``Compiled`` instance."""
-    from jax._src.interpreters import pxla
-
-    if isinstance(self._lowering, pxla.MeshComputation):
-      kw = dict(
-          _allow_propagation_to_outputs=[
-              pxla._is_unspecified(o)
-              for o in self._lowering.compile_args["out_shardings"]]
-      )
-    else:
-      kw = {}
-
+    kw: Dict[str, Any] = {"compiler_options": compiler_options}
     return Compiled(
-        self._lowering.compile(**kw),
+        self._lowering.compile(**kw),  # pytype: disable=wrong-keyword-args
         self.args_info,
         self.out_tree,
-        no_kwargs=self._no_kwargs)
+        no_kwargs=self._no_kwargs,
+    )
 
   def as_text(self, dialect: Optional[str] = None) -> str:
     """A human-readable text representation of this lowering.
