@@ -47,7 +47,7 @@ from jax._src.interpreters import pxla
 from jax._src.config import (flags, bool_env, config,
                              raise_persistent_cache_errors,
                              persistent_cache_min_compile_time_secs)
-from jax._src.numpy.util import _promote_dtypes, _promote_dtypes_inexact
+from jax._src.numpy.util import promote_dtypes, promote_dtypes_inexact
 from jax._src.util import unzip2
 from jax._src.public_test_util import (  # noqa: F401
     _assert_numpy_allclose, _check_dtypes_match, _default_tolerance, _dtype, check_close, check_grads,
@@ -178,7 +178,6 @@ def capture_stdout() -> Generator[Callable[[], str], None, None]:
 
 @contextmanager
 def count_device_put():
-  device_put = dispatch.device_put
   batched_device_put = pxla.batched_device_put
   count = [0]
 
@@ -189,24 +188,19 @@ def count_device_put():
       # underlying payload or several). We only want to count these
       # recursive puts once, so we skip counting more than the outermost
       # one in such a call stack.
-      dispatch.device_put = device_put
       pxla.batched_device_put = batched_device_put
       try:
         return fn(*args, **kwargs)
       finally:
-        dispatch.device_put = device_put_and_count
         pxla.batched_device_put = batched_device_put_and_count
     return fn_and_count
 
-  device_put_and_count = make_fn_and_count(device_put)
   batched_device_put_and_count = make_fn_and_count(batched_device_put)
 
-  dispatch.device_put = device_put_and_count
   pxla.batched_device_put = batched_device_put_and_count
   try:
     yield count
   finally:
-    dispatch.device_put = device_put
     pxla.batched_device_put = batched_device_put
 
 
@@ -222,7 +216,7 @@ def count_primitive_compiles():
 
 
 @contextmanager
-def count_pjit_cache_miss():
+def count_pjit_cpp_cache_miss():
   original_pjit_lower = pjit_lib._pjit_lower
   count = [0]
 
@@ -242,18 +236,18 @@ def count_jit_and_pmap_compiles():
   # No need to clear any caches since we generally jit and pmap fresh callables
   # in tests.
 
-  mlir_jaxpr_subcomp = mlir.jaxpr_subcomp
+  mlir_lower = mlir.lower_jaxpr_to_module
   count = [0]
 
-  def mlir_jaxpr_subcomp_and_count(*args, **kwargs):
+  def mlir_lower_and_count(*args, **kwargs):
     count[0] += 1
-    return mlir_jaxpr_subcomp(*args, **kwargs)
+    return mlir_lower(*args, **kwargs)
 
-  mlir.jaxpr_subcomp = mlir_jaxpr_subcomp_and_count
+  mlir.lower_jaxpr_to_module = mlir_lower_and_count
   try:
     yield count
   finally:
-    mlir.jaxpr_subcomp = mlir_jaxpr_subcomp
+    mlir.lower_jaxpr_to_module = mlir_lower
 
 @contextmanager
 def assert_num_jit_and_pmap_compilations(times):
@@ -297,6 +291,15 @@ def is_device_cuda():
 def is_cloud_tpu():
   return 'libtpu' in xla_bridge.get_backend().platform_version
 
+
+def is_se_tpu():
+  return (
+      is_cloud_tpu() and not xla_bridge.using_pjrt_c_api()
+  ) or xla_bridge.get_backend().platform_version.startswith(
+      'StreamExecutor TPU'
+  )
+
+
 def is_device_tpu_v4():
   return jax.devices()[0].device_kind == "TPU v4"
 
@@ -493,7 +496,7 @@ def rand_fullrange(rng, standardize_nans=False):
   """Random numbers that span the full range of available bits."""
   def gen(shape, dtype, post=lambda x: x):
     dtype = np.dtype(dtype)
-    size = dtype.itemsize * np.prod(_dims_of_shape(shape), dtype=int)
+    size = dtype.itemsize * math.prod(_dims_of_shape(shape))
     vals = rng.randint(0, np.iinfo(np.uint8).max, size=size, dtype=np.uint8)
     vals = post(vals).view(dtype)
     if shape is PYTHON_SCALAR_SHAPE:
@@ -827,7 +830,7 @@ def promote_like_jnp(fun, inexact=False):
   tests make an np reference implementation act more like an jnp
   implementation.
   """
-  _promote = _promote_dtypes_inexact if inexact else _promote_dtypes
+  _promote = promote_dtypes_inexact if inexact else promote_dtypes
   def wrapper(*args, **kw):
     flat_args, tree = tree_flatten(args)
     args = tree_unflatten(tree, _promote(*flat_args))
@@ -911,8 +914,8 @@ class JaxTestCase(parameterized.TestCase):
 
   def assertDtypesMatch(self, x, y, *, canonicalize_dtypes=True):
     if not config.x64_enabled and canonicalize_dtypes:
-      self.assertEqual(_dtypes.canonicalize_dtype(_dtype(x)),
-                       _dtypes.canonicalize_dtype(_dtype(y)))
+      self.assertEqual(_dtypes.canonicalize_dtype(_dtype(x), allow_opaque_dtype=True),
+                       _dtypes.canonicalize_dtype(_dtype(y), allow_opaque_dtype=True))
     else:
       self.assertEqual(_dtype(x), _dtype(y))
 
@@ -1022,18 +1025,13 @@ class JaxTestCase(parameterized.TestCase):
                         atol=atol or tol, rtol=rtol or tol,
                         canonicalize_dtypes=canonicalize_dtypes)
 
-_CPP_JIT_IMPLEMENTATION = functools.partial(api._jit, True)
-_CPP_JIT_IMPLEMENTATION._name = "cpp"
-_PYTHON_JIT_IMPLEMENTATION = functools.partial(api._jit, False)
-_PYTHON_JIT_IMPLEMENTATION._name = "python"
 _PJIT_IMPLEMENTATION = jax.jit
-_PJIT_IMPLEMENTATION._name = "pjit"
+_PJIT_IMPLEMENTATION._name = "jit"
 _NOOP_JIT_IMPLEMENTATION = lambda x, *args, **kwargs: x
 _NOOP_JIT_IMPLEMENTATION._name = "noop"
 
 JIT_IMPLEMENTATION = (
-  _CPP_JIT_IMPLEMENTATION,
-  _PYTHON_JIT_IMPLEMENTATION,
+  _PJIT_IMPLEMENTATION,
   _NOOP_JIT_IMPLEMENTATION,
 )
 
