@@ -263,22 +263,26 @@ no_initializer = object()
 
 @overload
 def tree_reduce(function: Callable[[T, Any], T],
-                tree: Any) -> T:
+                tree: Any,
+                *,
+                is_leaf: Optional[Callable[[Any], bool]] = None) -> T:
     ...
 
 @overload
 def tree_reduce(function: Callable[[T, Any], T],
                 tree: Any,
-                initializer: T) -> T:
+                initializer: T,
+                is_leaf: Optional[Callable[[Any], bool]] = None) -> T:
     ...
 
 def tree_reduce(function: Callable[[T, Any], T],
                 tree: Any,
-                initializer: Any = no_initializer) -> T:
+                initializer: Any = no_initializer,
+                is_leaf: Optional[Callable[[Any], bool]] = None) -> T:
   if initializer is no_initializer:
-    return functools.reduce(function, tree_leaves(tree))
+    return functools.reduce(function, tree_leaves(tree, is_leaf=is_leaf))
   else:
-    return functools.reduce(function, tree_leaves(tree), initializer)
+    return functools.reduce(function, tree_leaves(tree, is_leaf=is_leaf), initializer)
 
 def tree_all(tree: Any) -> bool:
   return all(tree_leaves(tree))
@@ -431,6 +435,22 @@ def prefix_errors(prefix_tree: Any, full_tree: Any,
 def equality_errors(
     tree1: Any, tree2: Any, is_leaf: Optional[Callable[[Any], bool]] = None,
 ) -> Iterable[Tuple[KeyPath, str, str, str]]:
+  """Helper to describe structural differences between two pytrees.
+
+  Args:
+    tree1, tree2: pytrees to compare.
+
+  Usage:
+
+    raise Exception(
+        "Value 1 and value 2 must have the same pytree structure, but they have "
+        "the following structural differences:\n" +
+        ("\n".join(
+           f"   - {keystr(path)} is a {thing1} in value 1 and a {thing2} in "
+           f" value 2, so {explanation}.\n"
+           for path, thing1, thing2, explanation
+           in equality_errors(val1, val2))))
+  """
   yield from _equality_errors((), tree1, tree2, is_leaf)
 
 # TODO(mattjj): maybe share some logic with _prefix_error?
@@ -545,6 +565,14 @@ KeyEntry = TypeVar("KeyEntry", bound=Hashable)
 KeyPath = Tuple[KeyEntry, ...]
 
 def keystr(keys: KeyPath):
+  """Helper to pretty-print a tuple of keys.
+
+  Args:
+    keys: A tuple of ``KeyEntry`` or any class that can be converted to string.
+
+  Returns:
+    A string that joins all string representations of the keys.
+  """
   return ''.join([str(k) for k in keys])
 
 
