@@ -20,21 +20,24 @@ XLA. There are also a handful of related casting utilities.
 """
 
 from functools import partial, lru_cache
+import importlib
 import io
 import json
 import logging
 import os
 import platform as py_platform
+import pkgutil
+import sys
 import threading
 from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union
 import warnings
-
 import numpy as np
 
 from jax._src import lib
 from jax._src import distributed
 from jax._src.config import flags, bool_env, config, int_env
 from jax._src.lib import xla_client
+from jax._src.lib import xla_extension_version
 from jax._src import traceback_util
 from jax._src import util
 
@@ -45,6 +48,17 @@ try:
 except (ModuleNotFoundError, ImportError):
   iree = None
 
+logger = logging.getLogger(__name__)
+
+jax_plugins: Optional[Any]
+try:
+  import jax_plugins  # type: ignore
+except ModuleNotFoundError:
+  jax_plugins = None
+except ImportError as e:
+  logger.error("Failed to import jax_plugins: %s", e)
+  jax_plugins = None
+
 traceback_util.register_exclusion(__file__)
 
 
@@ -52,7 +66,6 @@ XlaBackend = xla_client.Client
 
 FLAGS = flags.FLAGS
 
-logger = logging.getLogger(__name__)
 
 # TODO(phawkins): Remove jax_xla_backend.
 flags.DEFINE_string(
@@ -93,7 +106,9 @@ def get_compile_options(
     use_spmd_partitioning: bool = True,
     use_auto_spmd_partitioning: bool = False,
     auto_spmd_partitioning_mesh_shape=[],
-    auto_spmd_partitioning_mesh_ids=[]) -> xla_client.CompileOptions:
+    auto_spmd_partitioning_mesh_ids=[],
+    env_options_overrides: Optional[Dict[str, str]] = None,
+) -> xla_client.CompileOptions:
   """Returns the compile options to use, as derived from flag values.
 
   Args:
@@ -111,6 +126,7 @@ def get_compile_options(
       auto_spmd_partitioning search space.
     auto_spmd_partitioning_mesh_ids: device ids used to create
       auto_spmd_partitioning search space.
+    env_options_overrides: dict of additional options parsed by the compiler
   """
   compile_options = xla_client.CompileOptions()
   compile_options.num_replicas = num_replicas
@@ -147,6 +163,9 @@ def get_compile_options(
     assert device_assignment.computation_count() == num_partitions
     compile_options.device_assignment = device_assignment
 
+  if env_options_overrides is not None:
+    compile_options.env_option_overrides = list(env_options_overrides.items())
+
   debug_options = compile_options.executable_build_options.debug_options
   if lib.cuda_path is not None:
     debug_options.xla_gpu_cuda_data_dir = lib.cuda_path
@@ -216,11 +235,24 @@ def make_gpu_client(
   allowed_devices = None
   if visible_devices != "all":
     allowed_devices = {int(x) for x in visible_devices.split(",")}
-  return xla_client.make_gpu_client(
-    distributed_client=distributed.global_state.client,
-    node_id=distributed.global_state.process_id,
-    platform_name=platform_name,
-    allowed_devices=allowed_devices)
+
+  if xla_extension_version < 160:
+    return xla_client.make_gpu_client(
+        distributed_client=distributed.global_state.client,
+        node_id=distributed.global_state.process_id,
+        platform_name=platform_name,
+        allowed_devices=allowed_devices,
+    )
+  else:
+    # Remove `type: ignore` when the min jaxlib version (xla_extension_version)
+    # >= 160.
+    return xla_client.make_gpu_client(
+        distributed_client=distributed.global_state.client,
+        node_id=distributed.global_state.process_id,
+        num_nodes=distributed.global_state.num_processes,
+        platform_name=platform_name,
+        allowed_devices=allowed_devices,
+    )  # type: ignore
 
 
 if hasattr(xla_client, "make_gpu_client"):
@@ -294,8 +326,116 @@ def _get_pjrt_plugin_config(
     )
   return (config['library_path'], config.get('create_options'))
 
+def discover_pjrt_plugins() -> None:
+  """Discovers plugins in the namespace package `jax_plugins` and import them.
 
-def register_pjrt_plugin_factories(plugins_from_env: str) -> None:
+  There are two methods used to discover plugin modules. They are intended
+  to be used together by implementors in order to cover all packaging and
+  development cases:
+
+  1. Define a globally unique module under the `jax_plugins` namespace
+     package (i.e. just create a `jax_plugins` directory and define your
+     module below it).
+  2. If building a package via pyproject.toml or setup.py, advertise your
+     plugin module name by including an entry-point under the `jax_plugins`
+     group which points to your full module name.
+
+  During Jax startup, Jax will load each module discovered in such a way and
+  call its `initialize()` function. It is expected that this function should
+  register its concrete plugin name/implementations via call(s) to
+  `jax._src.xla_bridge.register_plugin(name, priority=, library_paty=,
+  options=)`. Since `initialize()` functions are called for all installed
+  plugins, they should avoid doing expensive, non-registration related work.
+
+  TODO: We should provide a variant of `register_plugin` which allows the
+  library_path and options to be resolved via a callback. This would enable
+  light-weight plugin registration in cases where options need to be derived
+  from heavy-weight system initialization.
+  """
+  plugin_modules = set()
+  # Scan installed modules under |jax_plugins|. Note that not all packaging
+  # scenarios are amenable to such scanning, so we also use the entry-point
+  # method to seed the list.
+  if jax_plugins:
+    for _, name, _ in pkgutil.iter_modules(
+        jax_plugins.__path__, jax_plugins.__name__ + '.'
+    ):
+      logger.debug("Discovered path based JAX plugin: %s", name)
+      plugin_modules.add(name)
+  else:
+    logger.debug("No jax_plugins namespace packages available")
+
+  # Augment with advertised entrypoints.
+  if sys.version_info < (3, 10):
+    # Use the backport library because it provides a forward-compatible
+    # implementation.
+    from importlib_metadata import entry_points
+  else:
+    from importlib.metadata import entry_points
+
+  for entry_point in entry_points(group="jax_plugins"):
+    logger.debug("Discovered entry-point based JAX plugin: %s",
+                 entry_point.value)
+    plugin_modules.add(entry_point.value)
+
+  # Now load and initialize them all.
+  for plugin_module_name in plugin_modules:
+    logger.debug("Loading plugin module %s", plugin_module_name)
+    plugin_module = None
+    try:
+      plugin_module = importlib.import_module(plugin_module_name)
+    except ModuleNotFoundError:
+      logger.warning("Jax plugin configuration error: Plugin module %s "
+                     "does not exist", plugin_module_name)
+    except ImportError:
+      logger.exception("Jax plugin configuration error: Plugin module %s "
+                       "could not be loaded")
+
+    if plugin_module:
+      try:
+        plugin_module.initialize()
+      except:
+        logger.exception("Jax plugin configuration error: Exception when "
+                         "calling %s.initialize()", plugin_module_name)
+
+
+# TODO(b/261345120): decide on a public name and expose a public method which is
+# an alias of this method.
+def register_plugin(
+    plugin_name: str,
+    *,
+    priority: int = 400,
+    library_path: Optional[str] = None,
+    options: Optional[Mapping[str, Union[str, int, List[int], float]]] = None,
+) -> None:
+  """Registers a backend factory for the PJRT plugin.
+
+  Args:
+    plugin_name: the name of the plugin.
+    priority: the priority this plugin should be registered in jax backends.
+      Default to be 400.
+    library_path: Optional. The full path to the .so file of the plugin.
+      Required when the plugin is dynamically linked.
+    options: Optional. It is used when creating a PJRT plugin client.
+  """
+  def factory():
+    # Plugin may already be statically linked in some configurations.
+    if not xla_client.pjrt_plugin_loaded(plugin_name):
+      if library_path is None:
+        raise ValueError(
+            'The library path is None when trying to dynamically load the'
+            ' plugin.'
+        )
+      xla_client.load_pjrt_plugin_dynamically(plugin_name, library_path)
+    return xla_client.make_c_api_client(plugin_name, options)
+
+  logger.debug(
+      'registering PJRT plugin %s from %s', plugin_name, library_path
+  )
+  register_backend_factory(plugin_name, factory, priority=priority)
+
+
+def register_pjrt_plugin_factories_from_env() -> None:
   """Registers backend factories for PJRT plugins.
 
   A backend factory will be registered for every PJRT plugin in the input
@@ -309,37 +449,27 @@ def register_pjrt_plugin_factories(plugins_from_env: str) -> None:
 
   TPU PJRT plugin will be loaded and registered separately in make_tpu_client.
   """
-
-  def make_factory(name: str, path: str):
-    def factory():
-      if path.endswith('.json'):
-        library_path, options = _get_pjrt_plugin_config(path)
-      else:
-        library_path = path
-        options = None
-
-      xla_client.load_pjrt_plugin_dynamically(name, library_path)
-      return xla_client.make_c_api_client(name, options)
-
-    return factory
-
-  pjrt_plugins = _get_pjrt_plugin_names_and_library_paths(plugins_from_env)
-  for plugin_name, library_path in pjrt_plugins.items():
+  pjrt_plugins = _get_pjrt_plugin_names_and_library_paths(
+      os.getenv('PJRT_NAMES_AND_LIBRARY_PATHS', '')
+  )
+  for plugin_name, path in pjrt_plugins.items():
+    if path.endswith('.json'):
+      library_path, options = _get_pjrt_plugin_config(path)
+    else:
+      library_path = path
+      options = None
     logger.debug(
         'registering PJRT plugin %s from %s', plugin_name, library_path
     )
-    # It is assumed that if a plugin is installed, then the user wants to use
-    # the plugin by default. Therefore, plugins get the highest priority.
-    # For a PJRT plugin, its plugin_name is the same as its platform_name.
-    register_backend_factory(
-        plugin_name, make_factory(plugin_name, library_path), priority=400
-    )
+    register_plugin(plugin_name, library_path=library_path, options=options)
 
 
-# The plugin names and paths are set in env var PJRT_NAMES_AND_LIBRARY_PATHS,
+# Plugins in the namespace package `jax_plugins` will be imported.
+discover_pjrt_plugins()
+# Registers plugins names and paths set in env var PJRT_NAMES_AND_LIBRARY_PATHS,
 # in the format of 'name1:path1,name2:path2' ('name1;path1,name2;path2' for
 # windows).
-register_pjrt_plugin_factories(os.getenv('PJRT_NAMES_AND_LIBRARY_PATHS', ''))
+register_pjrt_plugin_factories_from_env()
 
 if iree is not None:
   register_backend_factory("iree", iree.iree_client_factory, priority=-100)
@@ -467,7 +597,9 @@ def _clear_backends() -> None:
 def _init_backend(platform: str) -> xla_client.Client:
   factory, unused_priority = _backend_factories.get(platform, (None, None))
   if factory is None:
-    raise RuntimeError(f"Unknown backend '{platform}'")
+    raise RuntimeError(
+        f"Backend '{platform}' is not in the list of known backends: "
+        f"{list(_backend_factories.keys())}.")
 
   logger.debug("Initializing backend '%s'", platform)
   backend = factory()
@@ -588,7 +720,7 @@ def default_backend() -> str:
   """Returns the platform name of the default XLA backend."""
   return get_backend(None).platform
 
-
+@lru_cache
 def local_devices(process_index: Optional[int] = None,
                   backend: Optional[Union[str, xla_client.Client]] = None,
                   host_id: Optional[int] = None) -> List[xla_client.Device]:
@@ -646,6 +778,7 @@ def host_id(backend: Optional[Union[str, xla_client.Client]] = None) -> int:
   return process_index(backend)
 
 
+@lru_cache
 def process_count(
     backend: Optional[Union[str, xla_client.Client]] = None
 ) -> int:
@@ -674,3 +807,13 @@ def host_ids(
 
 def using_pjrt_c_api(backend=None):
   return "PJRT C API" in get_backend(backend).platform_version
+
+
+# TODO(parkers): Get rid of this in favor of a generic way to get topologies.
+def make_pjrt_tpu_topology(topology_name=None, **kwargs):
+  # TODO(b/261484192): Make a system for lazily loading libtpu.so and call
+  # that inside make_tfrt_tpu_c_api_device_topology.
+  get_backend()  # Properly initialize libtpu.so.
+  return xla_client.make_tfrt_tpu_c_api_device_topology(
+      topology_name, **kwargs
+  )
