@@ -14,10 +14,11 @@
 
 import jax
 import inspect
+from typing import Optional
 from jax._src import core
 from jax import tree_util
 from jax._src import linear_util as lu
-from jax.experimental import pjit
+from jax._src import sharding_impls
 from jax.errors import UnexpectedTracerError
 from jax._src import mesh as mesh_lib
 from jax._src.lib.mlir.dialects import hlo
@@ -44,15 +45,26 @@ def _resolve_kwargs(fun, args, kwargs):
 class _ShardingCallbackInfo:
 
   def __init__(self, propagate_user_sharding, partition, to_mesh_pspec_sharding,
-               infer_sharding_from_operands, module_context, mesh, static_args):
+      in_tree, out_tree, infer_sharding_from_operands, module_context, mesh,
+      static_args):
     self.propagate_user_sharding = propagate_user_sharding
     self.partition = partition
     self.to_mesh_pspec_sharding = to_mesh_pspec_sharding
+    self.in_tree = in_tree
+    self.out_tree = out_tree
     self.infer_sharding_from_operands = infer_sharding_from_operands
     self.module_context = module_context
     self.mesh = mesh
     self.static_args = static_args
 
+  def unflatten_arg_shapes(self, arg_shapes, arg_shardings):
+    return self.in_tree.unflatten(
+        [
+            _to_jax_sharded_shape(s, self.to_mesh_pspec_sharding(sharding))
+            for s, sharding in zip(arg_shapes, arg_shardings)
+        ]
+    )
+
 
 _sharding_callbacks = weakref.WeakValueDictionary()  # type: ignore
 
@@ -63,62 +75,124 @@ def _to_jax_shape(s):
   return core.ShapedArray(s.dimensions(), s.numpy_dtype())
 
 
-def _custom_partitioning_propagate_user_sharding(sharding, shape, backend_string):
+def _to_jax_sharded_shape(s, sharding):
+  return jax.ShapeDtypeStruct(
+      s.dimensions(), s.numpy_dtype(), sharding=sharding
+  )
+
+
+def _pack_result_sharding(shape, result_shardings):
+  if shape.is_tuple():
+    return xc.HloSharding.tuple_sharding(shape, result_shardings)
+  else:
+    return result_shardings[0]
+
+
+def _flatten_sharding(tree, shardings, shapes):
+  return [
+      _to_hlo_sharding(sharding, len(shape.dimensions()))
+      for sharding, shape in zip(
+          tree.flatten_up_to(shardings), shapes
+      )
+  ]
+
+
+def _custom_partitioning_propagate_user_sharding(user_sharding, shape,
+                                                 backend_string):
   info = _sharding_callbacks[backend_string]
   if info.propagate_user_sharding is None:
-    return sharding
-  return info.propagate_user_sharding(*info.static_args, sharding, shape)
+    return user_sharding
+  if shape.is_tuple():
+    user_shapes = shape.tuple_shapes()
+    user_shardings = user_sharding.tuple_elements()
+  else:
+    user_shapes = (shape,)
+    user_shardings = (user_sharding,)
+  user_shape = info.out_tree.unflatten(
+          [
+              _to_jax_sharded_shape(s, info.to_mesh_pspec_sharding(sharding))
+              for s, sharding in zip(user_shapes, user_shardings)
+          ]
+      )
+  result_sharding = info.propagate_user_sharding(*info.static_args, user_shape)
+  result_shardings = _flatten_sharding(
+      info.out_tree, result_sharding, user_shapes)
+  return _pack_result_sharding(shape, result_shardings)
+
+
+def _to_hlo_sharding(sharding, num_dimensions):
+  if not isinstance(sharding, jax.sharding.XLACompatibleSharding):
+    raise ValueError(
+        "Custom Partitioning rules must return XLACompatibleShardings."
+    )
+  return sharding._to_xla_hlo_sharding(num_dimensions)
 
 
 def _custom_partitioning_partition(arg_shapes, arg_shardings, result_shape,
                                    result_sharding, backend_string):
   info = _sharding_callbacks[backend_string]
+  if result_shape.is_tuple():
+    result_shapes = result_shape.tuple_shapes()
+    result_shardings = result_sharding.tuple_elements()
+  else:
+    result_shapes = (result_shape,)
+    result_shardings = (result_sharding,)
   lower_fn, result_sharding, arg_shardings = info.partition(
-      *info.static_args,
-      [_to_jax_shape(s) for s in arg_shapes],
-      [info.to_mesh_pspec_sharding(s.to_proto()) for s in arg_shardings],
-      _to_jax_shape(result_shape),
-      info.to_mesh_pspec_sharding(result_sharding.to_proto())
+      *info.static_args, info.unflatten_arg_shapes(arg_shapes, arg_shardings),
+      info.out_tree.unflatten(
+        [
+            _to_jax_sharded_shape(s, info.to_mesh_pspec_sharding(sharding))
+            for s, sharding in zip(result_shapes, result_shardings)
+        ]
+    )
   )
   module_context = info.module_context
 
-  def to_hlo_sharding(sharding, shape):
-    return xc.HloSharding.from_proto(
-        sharding._to_xla_op_sharding(len(shape.dimensions())))
-
-  result_sharding = to_hlo_sharding(result_sharding, result_shape)
-  arg_shardings = [
-      to_hlo_sharding(sharding, s)
-      for sharding, s in zip(arg_shardings, arg_shapes)
-  ]
+  result_shardings = _flatten_sharding(
+      info.out_tree, result_sharding, result_shapes)
+  arg_shardings = _flatten_sharding(info.in_tree, arg_shardings, arg_shapes)
   tiled_args = [
       _to_jax_shape(sharding.tile(s))
       for sharding, s in zip(arg_shardings, arg_shapes)
   ]
+  tiled_results = [
+      _to_jax_shape(sharding.tile(s))
+      for sharding, s in zip(result_shardings, result_shapes)
+  ]
   closed_jaxpr = jax.make_jaxpr(
       lower_fn, axis_env=list(info.mesh.shape.items()))(*tiled_args)
-  axis_context = mlir.SPMDAxisContext(info.mesh)
+  if closed_jaxpr.out_avals != tiled_results:
+    raise ValueError(
+        "Mismatch in result shapes. %s vs %s"
+        % (repr(closed_jaxpr.out_avals), repr(tiled_results))
+    )
+  axis_context = sharding_impls.SPMDAxisContext(info.mesh)
   built = mlir.build_xla_computation_helper(
       closed_jaxpr,
       name="tmp_xla_computation",
       platform=module_context.platform,
       backend_or_name=module_context.backend_or_name,
       axis_context=axis_context.extend_manual(frozenset(info.mesh.axis_names)))
+  result_sharding = _pack_result_sharding(result_shape, result_shardings)
   return built, arg_shardings, result_sharding
 
 
 def _custom_partitioning_infer_sharding_from_operands(arg_shapes, arg_shardings,
-                                                      shape, backend_string):
+                                                      result_shape,
+                                                      backend_string):
   info = _sharding_callbacks[backend_string]
-  result_shape = _to_jax_shape(shape)
-  result = info.infer_sharding_from_operands(
+  if result_shape.is_tuple():
+    result_shapes = result_shape.tuple_shapes()
+  else:
+    result_shapes = (result_shape,)
+  result_sharding = info.infer_sharding_from_operands(
       *info.static_args,
-      [_to_jax_shape(s) for s in arg_shapes],
-      [info.to_mesh_pspec_sharding(s.to_proto()) for s in arg_shardings],
-      result_shape
+      info.unflatten_arg_shapes(arg_shapes, arg_shardings),
+      info.out_tree.unflatten([_to_jax_shape(s) for s in result_shapes]),
   )
-  return xc.HloSharding.from_proto(
-      result._to_xla_op_sharding(len(result_shape.shape)))
+  result_shardings = _flatten_sharding(
+      info.out_tree, result_sharding, result_shapes)
+  return _pack_result_sharding(result_shape, result_shardings)
 
 
 custom_partitioning_p = core.Primitive("custom_partitioning")
@@ -128,16 +202,19 @@ custom_partitioning_p.multiple_results = True
 def _custom_partitioning_abstract_eval(*avals, call, in_tree, out_tree,
                                        propagate_user_sharding, partition,
                                        infer_sharding_from_operands,
+                                       decode_shardings,
                                        static_args):
   del in_tree, out_tree, propagate_user_sharding, partition
-  del infer_sharding_from_operands, static_args
+  del infer_sharding_from_operands, decode_shardings, static_args
   return call.out_avals
 
 
-def _custom_partitioning_impl(*args, call, in_tree, out_tree, propagate_user_sharding,
-                              partition, infer_sharding_from_operands, static_args):
+def _custom_partitioning_impl(*args, call, in_tree, out_tree,
+                              propagate_user_sharding,
+                              partition, infer_sharding_from_operands,
+                              decode_shardings, static_args):
   del in_tree, out_tree, propagate_user_sharding, partition
-  del infer_sharding_from_operands, static_args
+  del infer_sharding_from_operands, decode_shardings, static_args
   return core.jaxpr_as_fun(call)(*args)
 
 
@@ -166,18 +243,23 @@ class custom_partitioning:
     def f(*args):
       return ...
 
-    def propagate_user_sharding(sharding, shape):
-      '''Update the sharding of the op from a user's sharding.'''
+    def propagate_user_sharding(user_shape):
+      '''Update the sharding of the op from a user's shape.sharding.'''
+      user_sharding = jax.tree_map(lambda x: x.sharding, user_shape)
 
-    def partition(arg_shapes, arg_shardings, result_shape, result_sharding):
+    def partition(arg_shapes, result_shape):
       def lower_fn(*args):
         ... builds computation on per-device shapes ...
+      result_shardings = jax.tree_map(lambda x: x.sharding, result_shape)
+      arg_shardings = jax.tree_map(lambda x: x.sharding, arg_shapes)
       # result_sharding and arg_shardings may optionally be modified and the
       # partitioner will insert collectives to reshape.
       return lower_fn, result_sharding, arg_shardings
 
-    def infer_sharding_from_operands(arg_shapes, arg_shardings, shape):
+    def infer_sharding_from_operands(arg_shapes, shape):
       '''Compute the result sharding from the sharding of the operands.'''
+      arg_shardings = jax.tree_map(lambda x: x.sharding, arg_shapes)
+
 
     f.def_partition(partition, propagate_user_sharding, infer_sharding_from_operands)
 
@@ -192,6 +274,9 @@ class custom_partitioning:
     inputs to match).
   * ``infer_sharding_from_operands``: Callable which computes an output ``NamedSharding``
     from the ``NamedSharding`` chosen for each argument.
+  * ``decode_shardings``: When set to True, convert input ``GSPMDSharding``s to
+    ``NamedSharding`` if possible. This may not be possible if the user does not
+    provide a contextual mesh.
 
   Positional arguments can be specified as static using static_argnums. JAX uses
   :code:`inspect.signature(fun)` to resolve these positional arguments.
@@ -229,12 +314,15 @@ class custom_partitioning:
           names = tuple(sharding.spec[:max_shared_dims]) + tuple(None for _ in range(rank - max_shared_dims))
           return NamedSharding(sharding.mesh, P(*names))
 
-      def partition(arg_shapes, arg_shardings, result_shape, result_sharding):
+      def partition(arg_shapes, result_shape):
+          result_shardings = jax.tree_map(lambda x: x.sharding, result_shape)
+          arg_shardings = jax.tree_map(lambda x: x.sharding, arg_shapes)
           return fft, \
               supported_sharding(arg_shardings[0], arg_shapes[0]), \
-              [supported_sharding(arg_shardings[0], arg_shapes[0])]
+              (supported_sharding(arg_shardings[0], arg_shapes[0]),)
 
-      def infer_sharding_from_operands(arg_shapes, arg_shardings, shape):
+      def infer_sharding_from_operands(arg_shapes, result_shape):
+          arg_shardings = jax.tree_map(lambda x: x.sharding, arg_shapes)
           return supported_sharding(arg_shardings[0], arg_shapes[0])
 
       @custom_partitioning
@@ -320,10 +408,11 @@ class custom_partitioning:
   __getattr__ = custom_api_util.forward_attr
 
   def def_partition(self, partition, infer_sharding_from_operands,
-                    propagate_user_sharding=None):
+                    propagate_user_sharding=None, decode_shardings=True):
     self.partition = partition
     self.propagate_user_sharding = propagate_user_sharding
     self.infer_sharding_from_operands = infer_sharding_from_operands
+    self.decode_shardings = decode_shardings
     return partition
 
   def __call__(self, *args, **kwargs):
@@ -358,6 +447,7 @@ class custom_partitioning:
         partition=self.partition,
         propagate_user_sharding=self.propagate_user_sharding,
         infer_sharding_from_operands=self.infer_sharding_from_operands,
+        decode_shardings=self.decode_shardings,
         in_tree=in_tree,
         out_tree=out_tree(),
         static_args=static_args
@@ -369,13 +459,14 @@ def _custom_partitioning_lowering_rule(ctx: mlir.LoweringRuleContext, *values,
                                        call, in_tree, out_tree,
                                        propagate_user_sharding, partition,
                                        infer_sharding_from_operands,
+                                       decode_shardings,
                                        static_args):
   mesh = mesh_lib.thread_resources.env.physical_mesh
   axis_context = ctx.module_context.axis_context
 
-  if isinstance(axis_context, mlir.ShardingContext):
+  if isinstance(axis_context, sharding_impls.ShardingContext):
     devices = axis_context.device_assignment
-  elif isinstance(axis_context, mlir.SPMDAxisContext):
+  elif isinstance(axis_context, sharding_impls.SPMDAxisContext):
     devices = list(axis_context.mesh.devices.flat)
   else:
     devices = None
@@ -384,33 +475,29 @@ def _custom_partitioning_lowering_rule(ctx: mlir.LoweringRuleContext, *values,
     return mlir.lower_fun(
         core.jaxpr_as_fun(call), multiple_results=True)(ctx, *values)
 
-  def to_mesh_pspec_sharding(op_sharding: xc.OpSharding):
-    if mesh.empty:
+  def to_mesh_pspec_sharding(op_sharding: Optional[xc.OpSharding]):
+    if op_sharding is None:
+      return op_sharding
+    if mesh.empty or not decode_shardings:
       from jax._src.sharding_impls import GSPMDSharding
       assert devices is not None
-      return GSPMDSharding(devices, op_sharding)
-    pspec = pjit.parse_flatten_op_sharding(op_sharding,
-                                           mesh)[0].get_partition_spec()
+      return GSPMDSharding(devices, op_sharding.to_proto())
+    pspec = sharding_impls.parse_flatten_op_sharding(
+        op_sharding.to_proto(), mesh)[0].get_partition_spec()
     return jax.sharding.NamedSharding(mesh, pspec)
 
-  sharding_callback_info = _ShardingCallbackInfo(propagate_user_sharding, partition,
-                                                to_mesh_pspec_sharding,
-                                                infer_sharding_from_operands,
-                                                ctx.module_context, mesh, static_args)
+  sharding_callback_info = _ShardingCallbackInfo(propagate_user_sharding,
+      partition, to_mesh_pspec_sharding, in_tree, out_tree,
+      infer_sharding_from_operands, ctx.module_context, mesh, static_args)
   key = str(id(sharding_callback_info))
   _sharding_callbacks[key] = sharding_callback_info
   # We need to make sure `sharding_callback_info` is still alive when the SPMD
   # partitioner runs so we keep it alive by attaching it to the executable.
   ctx.module_context.add_keepalive(sharding_callback_info)
 
-  mlir_shapes = [mlir.aval_to_ir_types(s) for s in call.out_avals]
-  if len(mlir_shapes) == 1:
-    out_type = mlir_shapes[0]
-  else:
-    out_type = [ir.TupleType.get_tuple(mlir_shapes)]
-
+  result_types = [mlir.aval_to_ir_type(s) for s in call.out_avals]
   out = hlo.CustomCallOp(
-      out_type,
+      result_types,
       list(values),
       call_target_name=ir.StringAttr.get(_CUSTOM_PARTITIONING_CALL_NAME),
       has_side_effect=ir.BoolAttr.get(False),
@@ -419,14 +506,7 @@ def _custom_partitioning_lowering_rule(ctx: mlir.LoweringRuleContext, *values,
       backend_config=ir.StringAttr.get(key),
       operand_layouts=None,
       result_layouts=None)
-  if len(mlir_shapes) == 1:
-    return [out.result]
-  else:
-    return [
-        hlo.GetTupleElementOp(out, mlir.i32_attr(i)).result
-        for i in range(len(mlir_shapes))
-    ]
-
+  return out.results
 
 mlir.register_lowering(custom_partitioning_p,
                        _custom_partitioning_lowering_rule)
