@@ -16,6 +16,7 @@
 Specific JAX primitive conversion tests are in primitives_test."""
 import collections
 import contextlib
+import math
 import os
 import re
 from typing import Callable, Dict, Optional, Tuple
@@ -33,8 +34,10 @@ from jax import sharding
 from jax._src import core
 from jax._src import source_info_util
 from jax._src import test_util as jtu
+from jax._src.lib import xla_client
+from jax._src.lib.mlir.dialects import stablehlo
 import jax._src.xla_bridge
-from jax.config import config
+from jax import config
 from jax.experimental import jax2tf
 from jax.experimental.jax2tf import jax_export
 from jax.experimental.jax2tf.tests import tf_test_util
@@ -66,20 +69,25 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(sin_x, f_tf(x))
     self.assertAllClose(sin_x, tf.function(f_tf, autograph=False,
                                            jit_compile=True)(x))
-    # TODO: The following, with jit_compile=False, fails with
-    # native serialization because the tf.function() somehow executes the
-    # XlaCallModule op on CPU. This is despite the `with tf.device()`
-    # tf_preferred_device = (
-    #     tf.config.list_logical_devices("TPU") +
-    #     tf.config.list_logical_devices("GPU") +
-    #     tf.config.list_logical_devices())[0]
-    # logging.info("Running TF on %s", tf_preferred_device)
-    # with tf.device(tf_preferred_device):
-    #   self.assertAllClose(sin_x, tf.function(f_tf, autograph=False,
-    #                                         jit_compile=False)(x))
 
-    # self.assertAllClose(sin_x, tf.function(f_tf, autograph=False,
-    #                                       jit_compile=False)(x))
+    tf_preferred_device = (
+        tf.config.list_logical_devices("TPU")
+        + tf.config.list_logical_devices("GPU")
+        + tf.config.list_logical_devices()
+    )[0]
+    logging.info("Running TF on %s", tf_preferred_device)
+
+    # The following, with jit_compile=False, fails with native serialization
+    # because TF executes the function where it is instantiated (For example,
+    # XlaCallModule op on CPU). The workaround here is that we can
+    # wrap it and add device assignment inside the tf.function.
+    @tf.function(autograph=False, jit_compile=False)
+    def f_tf_wrapped(x):
+      with tf.device(tf_preferred_device.name):
+        return f_tf(x)
+
+    with tf.device(tf_preferred_device.name):
+      self.assertAllClose(sin_x, f_tf_wrapped(x))
 
   def test_basics(self):
     f_jax = lambda x: jnp.sin(jnp.cos(x))
@@ -313,7 +321,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(4., tape.gradient(v, y))
 
   @jtu.sample_product(with_function=[False, True])
-  def test_gradients_pytree(self, with_function=True):
+  def test_gradients_pytree(self, with_function=False):
     def f(xy: Tuple[float, float]) -> Dict[str, float]:
       x, y = xy
       return dict(one=x * x, two=x * y)
@@ -385,7 +393,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(grad_jax.a, grad_tf[0])
     self.assertAllClose(grad_jax.b, grad_tf[1])
 
-
   @jtu.sample_product(with_function=[False, True])
   def test_gradients_with_ordered_dict_input(self, with_function=True):
     def f(inputs):
@@ -496,7 +503,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     d_dx_jax = grad_g(x)
     d_dx_tf = jax2tf.convert(grad_g)(x)
     self.assertEqual(d_dx_jax.dtype, dtypes.float0)
-    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), np.int32),
+    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), np.bool_),
                         d_dx_tf.numpy())
 
     shape = (3, 4)
@@ -504,7 +511,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     d_dx_jax = grad_g(x)
     d_dx_tf = jax2tf.convert(grad_g)(x)
     self.assertEqual(d_dx_jax.dtype, dtypes.float0)
-    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), np.int32),
+    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), np.bool_),
                         d_dx_tf.numpy())
 
   @jtu.sample_product(with_function=[False, True])
@@ -593,8 +600,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         if k in expected_overrides:
           if expected_overrides[k] == "ZERO":
             e = np.zeros_like(w)
-          elif expected_overrides[k] == "ZERO_INT32":
-            e = np.zeros(np.shape(w), dtype=np.int32)
+          elif expected_overrides[k] == "ZERO_BOOL":
+            e = np.zeros(np.shape(w), dtype=np.bool_)
           elif expected_overrides[k] == "ONE":
             e = np.ones_like(w)
           else:
@@ -612,7 +619,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
           print(f"Failed at {k}")
           raise
 
-
     # compare_with_overrides(g_jax, {},
     #   bool_passthrough=np.zeros(state["bool_passthrough"].shape, dtype=dtypes.float0),
     #   bool_unused=np.zeros(state["bool_unused"].shape, dtype=dtypes.float0),
@@ -624,7 +630,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     #   int_unused=np.zeros(state["int_unused"].shape, dtype=dtypes.float0),
     #   int_used=np.zeros(state["int_used"].shape, dtype=dtypes.float0))
 
-
     # Now native TF gradients, only to test how native TF AD works
     _, (grad_tf_0,) = tf_test_util.ComputeTfValueAndGrad(
         jax_f, args, unconnected_gradients=tf.UnconnectedGradients.ZERO)
@@ -665,10 +670,10 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     grad_tf_vjp_jax, = tf_vjp_jax_fun(*args_vjp)
     compare_with_overrides(what=grad_tf_vjp_jax,
                            expected=grad_tf_0,
-                           bool_passthrough="ZERO_INT32",
-                           bool_unused="ZERO_INT32", bool_used="ZERO_INT32",
-                           int_passthrough="ZERO_INT32", int_unused="ZERO_INT32",
-                           int_used="ZERO_INT32")
+                           bool_passthrough="ZERO_BOOL",
+                           bool_unused="ZERO_BOOL", bool_used="ZERO_BOOL",
+                           int_passthrough="ZERO_BOOL", int_unused="ZERO_BOOL",
+                           int_used="ZERO_BOOL")
 
   def test_readme_gradient_int(self):
     x = np.array(2, dtype=np.int16)
@@ -693,14 +698,13 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
                           unconnected_gradients=tf.UnconnectedGradients.ZERO))
       # returns 0 with the same shape and dtype as x
 
-
   def test_convert_argument_non_callable_error(self):
     with self.assertRaisesRegex(TypeError, "Expected a callable value"):
       jax2tf.convert(5.)
 
   def test_convert_argument_non_tensor_error(self):
     with self.assertRaisesRegex(TypeError,
-                                "Argument.*should be NumPy array"):
+                                "Argument.*is not a valid JAX type"):
       jax2tf.convert(lambda x: x)(lambda y: y)
 
   def test_argument_eager_tensor(self):
@@ -857,18 +861,33 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     res_tf = jax2tf.convert(jax.jit(f_jax, keep_unused=False))(x, y_unused)
     self.assertAllClose(f_jax(x, None), res_tf)
 
-  def test_jit_unused_grad(self):
+  @parameterized.named_parameters(
+      dict(testcase_name=mode, mode=mode)
+      for mode in ("eager", "graph", "compiled"))
+  def test_jit_unused_grad(self, mode="eager"):
     def f_jax(x, y_unused):
       return x * np.float32(2.)
 
     x, y_unused = np.float32(5.), np.arange(7, dtype=np.int32)
+    res_jax = f_jax(x, y_unused)
     f_tf = jax2tf.convert(jax.jit(f_jax, keep_unused=False))
-    xv, y_unused_v = tf.Variable(x), tf.Variable(y_unused)
-    with tf.GradientTape() as tape:
-      res_tf = f_tf(xv, y_unused_v)
-      grad_tf_x, grad_tf_y = tape.gradient(res_tf, (xv, y_unused_v))
 
-    self.assertAllClose(f_jax(x, None), res_tf)
+    x_tf, y_unused_tf = tf.constant(x), tf.constant(y_unused)
+    def grad_tf(x, y_unused):
+      with tf.GradientTape() as tape:
+        tape.watch(x)
+        tape.watch(y_unused)
+        res_tf = f_tf(x, y_unused)
+        grad_tf_x, grad_tf_y = tape.gradient(res_tf, (x, y_unused))
+      return res_tf, grad_tf_x, grad_tf_y
+
+    if mode == "graph":
+      grad_tf = tf.function(grad_tf, autograph=False)
+    elif mode == "compiled":
+      grad_tf = tf.function(grad_tf, autograph=False, jit_compile=True)
+
+    res_tf, grad_tf_x, grad_tf_y = grad_tf(x_tf, y_unused_tf)
+    self.assertAllClose(res_jax, res_tf)
     self.assertAllClose(np.float32(2.), grad_tf_x)
     self.assertIsNone(grad_tf_y)
 
@@ -877,7 +896,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       return jax2tf.convert(jnp.sin)(y)  # Inner convert takes tracer args
     with self.assertRaisesRegex(
         ValueError, "convert must be used outside all JAX transformations"):
-      jax2tf.convert(outer)(np.ones((4, )))
+      jax2tf.convert(outer)(np.ones((4,), dtype=np.float32))
 
   def test_nested_convert_error_non_tracer(self):
     """The inner convert takes non-tracer arguments"""
@@ -1049,7 +1068,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
                         jnp.bfloat16(2.750))
 
   @jtu.sample_product(with_function=[False, True])
-  def test_kwargs(self, with_function=True):
+  def test_kwargs(self, with_function=False):
     # Re: https://github.com/google/jax/issues/6791
     def f_jax(*, x):
       return jnp.sum(x)
@@ -1078,7 +1097,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
                          np.full_like(x[1], fill_value=2.)),
                         (grad_tf[0].numpy(), grad_tf[1].numpy()))
 
-
   @jtu.skip_on_flag("jax2tf_default_native_serialization", True)
   def test_enable_xla(self):
     # Tests that enable_xla flag is properly scoped to a conversion.
@@ -1271,7 +1289,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       return carry
 
     shape = (3, 2)
-    x = np.arange(np.prod(shape), dtype=np.float32).reshape(shape)
+    x = np.arange(math.prod(shape), dtype=np.float32).reshape(shape)
 
     jax_comp = jax.xla_computation(f_while)(x)
     backend = jax._src.xla_bridge.get_backend()
@@ -1516,11 +1534,11 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         stack.enter_context(mesh)
       # Run the JAX native version, to check it works, and to fill caches.
       _ = func_to_convert(*args)
-      exported = jax_export.serialize_native(
+      exported = jax_export.export(
           func_to_convert,
-          [core.ShapedArray(a.shape, a.dtype) for a in args],
           lowering_platform='tpu',
-          strict_checks=True)
+          strict_checks=True
+      )(*(core.ShapedArray(a.shape, a.dtype) for a in args))
 
     if transform1 == "shard_map":
       self.assertIn("stablehlo.all_gather", str(exported.mlir_module))
@@ -1587,6 +1605,60 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
           r"The current platform .* is not among the platforms required by the module: \[TPU\]"):
         f_grad_tf(x_v)
 
+  def test_effects_error(self):
+    def f_jax(x):
+      jax.debug.print("{}", x)
+      return jnp.sin(x)
+
+    with self.assertRaisesRegex(NotImplementedError,
+                                "serialization of host_callbacks is not yet implemented"):
+      jax2tf.convert(f_jax, native_serialization=True)(np.float32(42.))
+
+    def f_ordered_jax(x):
+      jax.debug.print("{}", x, ordered=True)
+      return jnp.sin(x)
+
+    with self.assertRaisesRegex(NotImplementedError,
+                                "serialization of host_callbacks is not yet implemented"):
+      jax2tf.convert(f_ordered_jax, native_serialization=True)(np.float32(42.))
+
+  def test_tuple_args(self):
+    # On TPU if we have more than 2000 arguments, we pass them as a tuple.
+    # This is a compiler option, and should have no effect on lowering.
+    if jtu.device_under_test() != "tpu":
+      raise unittest.SkipTest("Test enabled on TPU only")
+    def f_jax(*many_args):
+      acc = 0.
+      for a in many_args:
+        acc += a
+      return acc
+
+    many_args = [np.float32(i) for i in range(2001)]
+    # Test that we do set lowered.compile_args[tuple_args]
+    lowered = jax.jit(f_jax).lower(*many_args)
+    self.assertTrue(lowered._lowering.compile_args["tuple_args"])
+    res = jax2tf.convert(f_jax, native_serialization=True)(*many_args)
+    self.assertAllClose(f_jax(*many_args), res)
+
+  def test_nested_convert(self):
+    # Test call sequence: convert -> call_tf -> convert.
+
+    @jax.jit
+    def f_jax(x):
+      return x + 1
+
+    inputs = np.ones((10), dtype=np.float32)
+
+    res = f_jax(inputs)
+
+    f_tf = jax2tf.convert(f_jax, native_serialization=True)
+    self.assertAllClose(res, f_tf(inputs))
+
+    f_jax_nested = jax2tf.call_tf(f_tf)
+    self.assertAllClose(res, f_jax_nested(inputs))
+
+    f_tf_nested = jax2tf.convert(f_jax_nested, native_serialization=True)
+    self.assertAllClose(res, f_tf_nested(inputs))
 
 def get_serialized_computation(
     f_jax: Callable,
@@ -1602,10 +1674,18 @@ def get_serialized_computation(
   ).lower(*args)
   else:
     lowered = jax.jit(f_jax, abstracted_axes=abstracted_axes).lower(*args)
-  stablehlo_module_text = mlir.module_to_string(lowered._lowering.stablehlo())
-  logging.info("Serialized ir.Module = %s", stablehlo_module_text)
-  return stablehlo_module_text, 3
-
+  mlir_module = lowered._lowering.stablehlo()
+  xla_call_module_version = 5
+  mlir_str = mlir.module_to_bytecode(mlir_module)
+  if stablehlo.get_api_version() < 4:
+    target_version = stablehlo.get_earliest_forward_compatible_version()
+  else:
+    # See comments next to the usage of stablehlo.get_minimum_version() in
+    # jax_export.py for an explanation how it works.
+    target_version = stablehlo.get_minimum_version()
+  mlir_module_serialized = xla_client._xla.mlir.serialize_portable_artifact(
+    mlir_str, target_version)
+  return mlir_module_serialized, xla_call_module_version
 
 class XlaCallModuleTest(tf_test_util.JaxToTfTestCase):
   """Unit tests for XlaCallModule. Will move these eventually to TF."""
@@ -1673,7 +1753,7 @@ class XlaCallModuleTest(tf_test_util.JaxToTfTestCase):
       return x + y
 
     shape = (8, 10)
-    x = np.arange(np.prod(shape), dtype=np.float32).reshape(shape)
+    x = np.arange(math.prod(shape), dtype=np.float32).reshape(shape)
     in_axis_resources = (P("x"), P("x"))
     out_axis_resources = None
     res_jax = pjit.pjit(
@@ -1700,6 +1780,35 @@ class XlaCallModuleTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(res_tf.numpy(), res_jax)
 
 
+@jtu.with_config(jax_enable_custom_prng=True)
+class Jax2tfWithCustomPRNGTest(tf_test_util.JaxToTfTestCase):
+  def test_key_argument(self):
+    func = lambda key: jax.random.uniform(key, ())
+    key = jax.random.PRNGKey(0)
+    key_raw = jax.random.key_data(key)
+    with self.assertWarnsRegex(FutureWarning, "Raw arrays as random keys.*"):
+      tf_result = jax2tf.convert(func)(key_raw)
+    jax_result = func(key)
+    self.assertEqual(tf_result, jax_result)
+
+  def test_key_from_seed(self):
+    func = lambda seed: jax.random.uniform(jax.random.PRNGKey(seed), ())
+    seed = 1701
+    tf_result = jax2tf.convert(func)(seed)
+    jax_result = func(seed)
+    self.assertEqual(tf_result, jax_result)
+
+  def test_key_closure(self):
+    def func():
+      # Include nontrivial shape operations to catch tracing bugs.
+      key = global_key.reshape(1).squeeze()
+      return jax.random.uniform(key)
+    global_key = jax.random.PRNGKey(0)
+    tf_result = jax2tf.convert(func)()
+    jax_result = func()
+    self.assertEqual(tf_result, jax_result)
+
+
 if __name__ == "__main__":
   # TODO: Remove once tensorflow is 2.10.0 everywhere.
   if not hasattr(tfxla, "optimization_barrier"):
