@@ -28,7 +28,7 @@ from jax import numpy as jnp
 from jax._src import test_util as jtu
 from jax import tree_util
 
-from jax.config import config
+from jax import config
 from jax.experimental import jax2tf
 from jax._src import xla_bridge
 import numpy as np
@@ -48,14 +48,15 @@ def _run_tf_function(func_tf: Callable, *tf_args, mode: str):
   if mode == "eager":
     return func_tf(*tf_args)  # EAGER
   elif mode == "graph":
-    return tf.function(
+    return tf.function(  # GRAPH
         func_tf,
         autograph=False,
+        # Note that jit_compile defaults to True on TPU and False elsewhere
         input_signature=_make_tf_input_signature(*tf_args))(*tf_args)  # GRAPH
   elif mode == "compiled":
     # Adding an explicit input_signature prevents TF from constant-folding
     # the computation eagerly before compilation
-    return tf.function(
+    return tf.function(  # COMPILED
         func_tf,
         autograph=False,
         jit_compile=True,
@@ -215,6 +216,8 @@ class JaxToTfTestCase(jtu.JaxTestCase):
     unexpected_successes: List[str] = []
     # Run the "compiled" mode first, it is most important
     for mode in ("compiled", "eager", "graph"):
+      if mode == "graph" and jtu.device_under_test() == "tpu":
+        continue  # The "graph" mode on TPU is the same as "compiled"
       def log_message(extra):
         return f"[{self._testMethodName}] {mode=}: {extra}"
 
@@ -369,7 +372,7 @@ class JaxToTfTestCase(jtu.JaxTestCase):
 
   def TfToHlo(self, tf_fun: Callable, *args):
     # Converts a tf.function to HLO text which we can inspect for occurrence of
-    # substrings. This works whether we use native lowering or not.
+    # substrings. This works whether we use native serialization or not.
     tf_function = tf.function(tf_fun, autograph=False, jit_compile=True)
     device_name = f"/device:{jtu.device_under_test().upper()}:0"
     return tf_function.experimental_get_compiler_ir(*args)(stage="hlo",
@@ -381,7 +384,7 @@ class JaxToTfTestCase(jtu.JaxTestCase):
     # graph. We count the number of characters in the textual representation
     # of the constant.
     f_tf_graph = tf.function(tf_fun, autograph=False).get_concrete_function(*args).graph.as_graph_def()
-    if config.jax2tf_default_experimental_native_lowering:
+    if config.jax2tf_default_native_serialization:
       # This way of finding constants may be brittle, if the constant representation
       # contains >. It seems tobe hex-encoded, so this may be safe.
       large_consts = [m for m in re.findall(r"dense<([^>]+)>", str(f_tf_graph)) if len(m) >= at_least]
