@@ -53,18 +53,16 @@ from typing import (
 
 import numpy as np
 
-import jax
 from jax import lax
 from jax._src import core
 from jax._src import linear_util as lu
 from jax._src import pjit
+from jax._src import sharding_impls
 from jax.experimental.sparse.bcoo import bcoo_multiply_dense, bcoo_multiply_sparse
 import jax.numpy as jnp
 from jax._src.api_util import flatten_fun_nokwargs
 from jax._src.lib import pytree
 from jax._src.interpreters import partial_eval as pe
-from jax._src.interpreters import xla
-from jax._src.interpreters import pxla
 from jax.tree_util import tree_flatten, tree_map, tree_unflatten
 from jax.util import safe_map, safe_zip, split_list
 from jax._src.config import config
@@ -603,6 +601,7 @@ sparse_rules_bcoo[lax.transpose_p] = _transpose_sparse
 
 def _add_sparse(spenv, *spvalues):
   X, Y = spvalues
+  out_shape = lax.broadcast_shapes(X.shape, Y.shape)
   if X.is_sparse() and Y.is_sparse():
     if X.shape != Y.shape:
       raise NotImplementedError("Addition between sparse matrices of different shapes.")
@@ -621,7 +620,18 @@ def _add_sparse(spenv, *spvalues):
       out_data = lax.concatenate([spenv.data(X), spenv.data(Y)], dimension=spenv.indices(X).ndim - 2)
       out_spvalue = spenv.sparse(X.shape, out_data, out_indices)
   else:
-    raise NotImplementedError("Addition between sparse and dense array.")
+    if Y.is_sparse():
+      X, Y = Y, X
+    assert X.is_sparse() and Y.is_dense()
+    if Y.shape != out_shape:
+      raise NotImplementedError(
+        "Addition between a sparse array X and a dense array Y is not implemented when "
+        "the output shape is larger than Y.shape. This is to prevent silent densification "
+        "of a large sparse array. If this is your intent, you can explicitly cast the sparse "
+        "array to a dense matrix.")
+    X_promoted, Y_promoted = spvalues_to_arrays(spenv, (X, Y))
+    out = X_promoted.todense() + Y_promoted
+    out_spvalue = spenv.dense(out)
 
   return (out_spvalue,)
 
@@ -765,9 +775,13 @@ def _pjit_sparse(spenv, *spvalues, jaxpr, in_shardings, out_shardings,
   # TODO(yashkatariya, vanderplas): Flatten twice and set the correct sharding
   # for data and indices.
   in_shardings = in_shardings + tuple(
-      pxla._UNSPECIFIED for _ in range(len(args_flat) - len(in_shardings)))
+      sharding_impls.UNSPECIFIED
+      for _ in range(len(args_flat) - len(in_shardings))
+  )
   out_shardings = out_shardings + tuple(
-      pxla._UNSPECIFIED for _ in range(len(sp_call_jaxpr.out_avals) - len(out_shardings)))
+      sharding_impls.UNSPECIFIED
+      for _ in range(len(sp_call_jaxpr.out_avals) - len(out_shardings))
+  )
 
   out_flat = pjit.pjit_p.bind(
       *args_flat,
