@@ -16,27 +16,21 @@ from jax._src.interpreters.mlir import (
   AxisContext as AxisContext,
   ConstantHandler as ConstantHandler,
   DEVICE_TO_DEVICE_TYPE as DEVICE_TO_DEVICE_TYPE,
-  DimExprEvaluator as DimExprEvaluator,
   LoweringResult as LoweringResult,
   LoweringRule as LoweringRule,
   LoweringRuleContext as LoweringRuleContext,
-  Mesh as Mesh,
-  MeshAxisName as MeshAxisName,
   ModuleContext as ModuleContext,
   RECV_FROM_HOST_TYPE as RECV_FROM_HOST_TYPE,
-  ReplicaAxisContext as ReplicaAxisContext,
   SEND_TO_HOST_TYPE as SEND_TO_HOST_TYPE,
-  SPMDAxisContext as SPMDAxisContext,
-  ShardingContext as ShardingContext,
   Token as Token,
   TokenSet as TokenSet,
   Value as Value,
   _call_lowering as _call_lowering,
   _lowerings as _lowerings,
   _platform_specific_lowerings as _platform_specific_lowerings,
-  _xla_call_lower as _xla_call_lower,
   aval_to_ir_type as aval_to_ir_type,
   aval_to_ir_types as aval_to_ir_types,
+  core_call_lowering as core_call_lowering,
   dense_bool_elements as dense_bool_elements,
   dense_int_elements as dense_int_elements,
   dtype_to_ir_type as dtype_to_ir_type,
@@ -65,3 +59,11 @@ from jax._src.interpreters.mlir import (
   token_type as token_type,
   xla_computation_to_mlir_module as xla_computation_to_mlir_module,
 )
+
+from jax._src.mesh import Mesh as Mesh
+from jax._src.sharding_impls import (
+  MeshAxisName as MeshAxisName,
+  ReplicaAxisContext as ReplicaAxisContext,
+  SPMDAxisContext as SPMDAxisContext,
+  ShardingContext as ShardingContext,
+)
