@@ -13,13 +13,7 @@
 # limitations under the License.
 
 from jax._src.interpreters.pxla import (
-  AUTO as AUTO,
-  ArrayMapping as ArrayMapping,
-  ArrayMappingOrAutoOrUnspecified as ArrayMappingOrAutoOrUnspecified,
   AvalDimSharding as AvalDimSharding,
-  Chunked as Chunked,
-  DynamicAxisEnv as DynamicAxisEnv,
-  DynamicAxisEnvFrame as DynamicAxisEnvFrame,
   EmapInfo as EmapInfo,
   ExecuteReplicated as ExecuteReplicated,
   Index as Index,
@@ -29,51 +23,31 @@ from jax._src.interpreters.pxla import (
   MeshComputation as MeshComputation,
   MeshDimAssignment as MeshDimAssignment,
   MeshExecutable as MeshExecutable,
-  NoSharding as NoSharding,
-  OpShardingType as OpShardingType,
-  OrderedDictType as OrderedDictType,
   ParallelCallableInfo as ParallelCallableInfo,
-  PartitionInfo as PartitionInfo,
-  PartitionsOrReplicated as PartitionsOrReplicated,
   PmapComputation as PmapComputation,
   PmapExecutable as PmapExecutable,
   PxlaResultHandler as PxlaResultHandler,
   ReplicaInfo as ReplicaInfo,
-  Replicated as Replicated,
   ResultsHandler as ResultsHandler,
   SPMDBatchTrace as SPMDBatchTrace,
   ShardInfo as ShardInfo,
-  ShardedAxis as ShardedAxis,
-  ShardingSpec as ShardingSpec,
   TileManual as TileManual,
   TileVectorize as TileVectorize,
   TilingMethod as TilingMethod,
   UnloadedMeshExecutable as UnloadedMeshExecutable,
   UnloadedPmapExecutable as UnloadedPmapExecutable,
-  Unstacked as Unstacked,
   WeakRefList as WeakRefList,
-  _UNSPECIFIED as _UNSPECIFIED,
   _create_pmap_sharding_spec as _create_pmap_sharding_spec,
   _get_and_check_device_assignment as _get_and_check_device_assignment,
-  _get_sharding_specs as _get_sharding_specs,
-  _is_unspecified as _is_unspecified,
-  _one_replica_buffer_indices as _one_replica_buffer_indices,
   _pmap_sharding_spec as _pmap_sharding_spec,
-  are_op_shardings_equal as are_op_shardings_equal,
-  array_mapping_to_axis_resources as array_mapping_to_axis_resources,
   array_types as array_types,
   custom_resource_typing_rules as custom_resource_typing_rules,
   device_put as _deprecated_device_put,
-  find_partitions as find_partitions,
   find_replicas as find_replicas,
   full_to_shard_p as full_to_shard_p,
-  get_global_aval as get_global_aval,
-  get_local_aval as get_local_aval,
-  get_num_partitions as get_num_partitions,
   global_aval_to_result_handler as global_aval_to_result_handler,
   global_avals_to_results_handler as global_avals_to_results_handler,
   global_result_handlers as global_result_handlers,
-  is_op_sharding_replicated as is_op_sharding_replicated,
   local_aval_to_result_handler as local_aval_to_result_handler,
   local_avals_to_results_handler as local_avals_to_results_handler,
   local_result_handlers as local_result_handlers,
@@ -83,11 +57,7 @@ from jax._src.interpreters.pxla import (
   maybe_extend_axis_env as maybe_extend_axis_env,
   mesh_sharding_specs as mesh_sharding_specs,
   multi_host_supported_collectives as multi_host_supported_collectives,
-  new_mesh_sharding_specs as new_mesh_sharding_specs,
-  op_sharding_to_indices as op_sharding_to_indices,
   parallel_callable as parallel_callable,
-  partitioned_sharding_spec as partitioned_sharding_spec,
-  reconcile_num_partitions as reconcile_num_partitions,
   replicate as replicate,
   resource_typecheck as resource_typecheck,
   shard_arg_handlers as shard_arg_handlers,
@@ -96,8 +66,6 @@ from jax._src.interpreters.pxla import (
   shard_aval as shard_aval,
   shard_aval_handlers as shard_aval_handlers,
   shard_to_full_p as shard_to_full_p,
-  sharding_spec_sharding_proto as sharding_spec_sharding_proto,
-  spec_to_indices as spec_to_indices,
   spmd_primitive_batchers as spmd_primitive_batchers,
   stage_parallel_callable as stage_parallel_callable,
   tile_aval_nd as tile_aval_nd,
@@ -116,29 +84,42 @@ from jax._src.mesh import (
   thread_resources as thread_resources,
 )
 
+from jax._src.op_shardings import (
+  are_op_shardings_equal as are_op_shardings_equal,
+  is_op_sharding_replicated as is_op_sharding_replicated,
+  op_sharding_to_indices as op_sharding_to_indices,
+)
+
+from jax._src.sharding_impls import (
+  ArrayMapping as ArrayMapping,
+  ArrayMappingOrAutoOrUnspecified as ArrayMappingOrAutoOrUnspecified,
+  AUTO as AUTO,
+  UNSPECIFIED as _UNSPECIFIED,
+  array_mapping_to_axis_resources as array_mapping_to_axis_resources,
+  is_unspecified as _is_unspecified,
+)
+
+from jax._src.sharding_specs import (
+  Chunked as Chunked,
+  NoSharding as NoSharding,
+  OpShardingType as OpShardingType,
+  Replicated as Replicated,
+  ShardedAxis as ShardedAxis,
+  ShardingSpec as ShardingSpec,
+  Unstacked as Unstacked,
+  new_mesh_sharding_specs as new_mesh_sharding_specs,
+  sharding_spec_sharding_proto as sharding_spec_sharding_proto,
+  spec_to_indices as spec_to_indices,
+)
+
 # Deprecations
 
-from jax._src.mesh import Mesh as _deprecated_Mesh
 from jax._src.interpreters.pxla import (
-  PartitionSpec as _deprecated_PartitionSpec,
   ShardedDeviceArray as _deprecated_ShardedDeviceArray,
   make_sharded_device_array as _deprecated_make_sharded_device_array,
 )
 
 _deprecations = {
-    # Added Feb 8, 2023:
-    "Mesh": (
-        "jax.interpreters.pxla.Mesh is deprecated. Use jax.sharding.Mesh.",
-        _deprecated_Mesh,
-    ),
-    # Added Feb 8, 2023:
-    "PartitionSpec": (
-        (
-            "jax.interpreters.pxla.PartitionSpec is deprecated. Use "
-            "jax.sharding.PartitionSpec."
-        ),
-        _deprecated_PartitionSpec,
-    ),
     # Added March 15, 2023:
     "ShardedDeviceArray": (
         (
@@ -167,9 +148,7 @@ _deprecations = {
 
 import typing
 if typing.TYPE_CHECKING:
-  from jax._src.mesh import Mesh as Mesh
   from jax._src.interpreters.pxla import (
-    PartitionSpec as PartitionSpec,
     ShardedDeviceArray as ShardedDeviceArray,
     device_put as device_put,
     make_sharded_device_array as make_sharded_device_array,
