@@ -142,10 +142,10 @@ from jax._src.numpy.lax_numpy import (
     inf as inf,
     inner as inner,
     insert as insert,
+    int8 as int8,
     int16 as int16,
     int32 as int32,
     int64 as int64,
-    int8 as int8,
     int_ as int_,
     integer as integer,
     interp as interp,
@@ -168,9 +168,9 @@ from jax._src.numpy.lax_numpy import (
     logspace as logspace,
     mask_indices as mask_indices,
     matmul as matmul,
+    matrix_transpose as matrix_transpose,
     meshgrid as meshgrid,
     moveaxis as moveaxis,
-    msort as msort,
     nan as nan,
     nan_to_num as nan_to_num,
     nanargmax as nanargmax,
@@ -237,10 +237,10 @@ from jax._src.numpy.lax_numpy import (
     triu_indices_from as triu_indices_from,
     trunc as trunc,
     uint as uint,
+    uint8 as uint8,
     uint16 as uint16,
     uint32 as uint32,
     uint64 as uint64,
-    uint8 as uint8,
     unpackbits as unpackbits,
     unravel_index as unravel_index,
     unsignedinteger as unsignedinteger,
@@ -254,6 +254,22 @@ from jax._src.numpy.lax_numpy import (
     zeros_like as zeros_like,
 )
 
+# TODO(phawkins): make this import unconditional after increasing the ml_dtypes
+# minimum version.
+import jax._src.numpy.lax_numpy
+if hasattr(jax._src.numpy.lax_numpy, "int4"):
+  from jax._src.numpy.lax_numpy import (
+    int4 as int4,
+    uint4 as uint4,
+  )
+# TODO(jakevdp): make this import unconditional after increasing the minimum
+# version for ml_dtypes and jaxlib
+if hasattr(jax._src.numpy.lax_numpy, "float8_e4m3b11fnuz"):
+  from jax._src.numpy.lax_numpy import (
+    float8_e4m3b11fnuz as float8_e4m3b11fnuz,
+  )
+
+
 from jax._src.numpy.index_tricks import (
   c_ as c_,
   index_exp as index_exp,
@@ -277,7 +293,6 @@ from jax._src.numpy.polynomial import (
 )
 
 from jax._src.numpy.reductions import (
-    alltrue as alltrue,
     amin as amin,
     amax as amax,
     any as any,
@@ -286,7 +301,6 @@ from jax._src.numpy.reductions import (
     count_nonzero as count_nonzero,
     cumsum as cumsum,
     cumprod as cumprod,
-    cumproduct as cumproduct,
     max as max,
     mean as mean,
     median as median,
@@ -305,10 +319,8 @@ from jax._src.numpy.reductions import (
     nanvar as nanvar,
     percentile as percentile,
     prod as prod,
-    product as product,
     ptp as ptp,
     quantile as quantile,
-    sometrue as sometrue,
     std as std,
     sum as sum,
     var as var,
@@ -431,11 +443,32 @@ _deprecations = {
         "jax.numpy.DeviceArray is deprecated. Use jax.Array.",
         ndarray,
     ),
+    # Added June 2, 2023:
+    "alltrue": (
+        "jax.numpy.alltrue is deprecated. Use jax.numpy.all",
+        all,
+    ),
+    "cumproduct": (
+        "jax.numpy.cumproduct is deprecated. Use jax.numpy.cumprod",
+        cumprod,
+    ),
+    "product": (
+        "jax.numpy.product is deprecated. Use jax.numpy.prod",
+        prod,
+    ),
+    "sometrue": (
+        "jax.numpy.sometrue is deprecated. Use jax.numpy.any",
+        any,
+    ),
 }
 
 import typing
 if typing.TYPE_CHECKING:
   from jax._src.basearray import Array as DeviceArray
+  alltrue = all
+  cumproduct = cumprod
+  product = prod
+  sometrue = any
 else:
   from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
   __getattr__ = _deprecation_getattr(__name__, _deprecations)
