@@ -39,19 +39,3 @@ from jax._src.third_party.numpy.linalg import (
   tensorinv as tensorinv,
   tensorsolve as tensorsolve,
 )
-
-# Module initialization is encapsulated in a function to avoid accidental
-# namespace pollution.
-_NOT_IMPLEMENTED = []
-def _init():
-  import numpy as np
-  from jax._src.numpy import lax_numpy
-  from jax._src import util
-  # Builds a set of all unimplemented NumPy functions.
-  for name, func in util.get_module_functions(np.linalg).items():
-    if name not in globals():
-      _NOT_IMPLEMENTED.append(name)
-      globals()[name] = lax_numpy._not_implemented(func)
-
-_init()
-del _init
